"""
This submodule implements different pooling modules.

.. autoclass:: tamm.layers.SimpleAdaptiveAvgPooling
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.ConvPooler
    :members: create_basic_builder
"""

from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm.layers import activation as _activation
from tamm.layers import basic as _basic
from tamm.layers import convolution as _convolution
from tamm.layers import embedding as _embedding
from tamm.layers import feed_forward as _feed_forward
from tamm.layers import norm as _norm
from tamm.layers import sequential as _sequential
from tamm.layers import torch_nn as _torch_nn
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.positional_encoding import (
    SpatialPositionalEmbedding as _SpatialPositionalEmbedding,
)
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class SimpleAdaptiveAvgPooler(_nn.Module, _LayerMixin):
    """
    Applies average pooling to produce outputs with a specific shape.
    For the pooling, this layer uses ``stride = input_size // output_size`` and
    ``kernel_size = input_size - (output_size - 1) * stride``.  While this is a
    common way to perform adaptive average pooling, note that it may differ
    from :class:`torch.nn.AdaptiveAvgPool2d`.

    This layer expects inputs in NCHW format.

    Args:
        output_shape (:obj:`tuple` or :obj:`int`): The target height and width of the
            output.  This can be a single :obj:`int` for a square.  If either height
            or width are ``None``, then the output dimension is the same as the input
            dimension.
    """

    def __init__(self, output_shape: _Union[int, _Tuple[int, int]]):
        super().__init__()
        try:
            height, width = output_shape
        except TypeError:
            height, width = (output_shape, output_shape)

        if not isinstance(height, (int, type(None))):
            raise TypeError(f"height must be an integer or None, not a {type(height)}")
        if not isinstance(width, (int, type(None))):
            raise TypeError(f"width must be an integer or None, not a {type(width)}")

        self.output_shape = (height, width)

    def extra_repr(self):
        return f"output_shape={self.output_shape}"

    def forward(self, x: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            x (:obj:`torch.Tensor`): The input tensor in NCHW format.
        """

        input_shape = x.shape[-2:]
        output_shape = tuple(
            odim if odim is not None else idim
            for idim, odim in zip(input_shape, self.output_shape)
        )

        stride = tuple(idim // odim for idim, odim in zip(input_shape, output_shape))
        kernel_size = tuple(
            idim - (odim - 1) * stride_i
            for idim, odim, stride_i in zip(input_shape, output_shape, stride)
        )

        output = _torch.nn.functional.avg_pool2d(  # pylint: disable=not-callable
            x, kernel_size=kernel_size, stride=stride, padding=0
        )

        true_output_shape = output.shape[-2:]
        if true_output_shape != output_shape:
            raise ValueError(
                f"Output shape is not as expected ({true_output_shape} != "
                f"{output_shape}) "
            )

        return output


class CAbstractorPooler(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    An image pooler from the paper `Honeybee: Locality-enhanced Projector for
    Multimodal LLM <https://arxiv.org/abs/2312.06742>`__.  This layer transforms
    2D embeddings using a convolution block, 2D pooler, and then another
    convolution block.  It also contains various layers for extra tokens, positional
    encodings, and feed forward passes.

    This layer expects inputs in channels-last format.
    """

    def __init__(
        self,
        *,
        drop_tokens: _OptionalModuleOrBuilder = None,
        to_channels_first: _ModuleOrBuilder,
        positional_encoding: _OptionalModuleOrBuilder = None,
        convolution_0: _OptionalModuleOrBuilder = None,
        pooler: _ModuleOrBuilder,
        convolution_1: _OptionalModuleOrBuilder = None,
        to_channels_last: _ModuleOrBuilder,
        feed_forward: _OptionalModuleOrBuilder = None,
        add_tokens: _OptionalModuleOrBuilder = None,
    ):
        layers = {
            "drop_tokens": drop_tokens,
            "to_channels_first": to_channels_first,
            "positional_encoding": positional_encoding,
            "convolution_0": convolution_0,
            "pooler": pooler,
            "convolution_1": convolution_1,
            "to_channels_last": to_channels_last,
            "feed_forward": feed_forward,
            "add_tokens": add_tokens,
        }
        super().__init__(layers)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        input_dim: int,
        convolution_output_dim: _Optional[int] = None,
        feed_forward_hidden_dim: _Optional[int] = None,
        output_dim: _Optional[int] = None,
        apply_positional_embedding: bool = False,
        positional_embedding_shape: _Optional[_Tuple[int, int]] = None,
        positional_embedding_interpolation: _Optional[str] = None,
        blocks_per_regnet_stage: int = 3,
        regnet_activation: str = "relu",
        norm_eps: float = 1e-5,
        norm_bias: bool = False,
        squeeze_excitation_reduced_dim: _Optional[int] = None,
        squeeze_excitation_bias: bool = False,
        pooler_output_shape: _Tuple[int, int],
        feed_forward_activation: str = "relu",
        num_prepend_tokens: int = 0,
        num_append_tokens: int = 0,
    ):
        """
        Creates a builder for a :obj:`.CAbstractorPooler`.

        Args:
            input_dim (:obj:`int`): The number of features in the inputs.
            convolution_output_dim (:obj:`int`, optional): The output dimension of the
                convolution stages.  If ``None``, this defaults to ``input_dim``.
            feed_forward_hidden_dim (:obj:`int`, optional): The hidden dimension of the
                feed forward layer.  If ``None``, this defaults to
                ``convolution_output_dim``.
            output_dim (:obj:`int`, optional): The output dimension of the layer.  If
                ``None``, this defaults to ``input_dim``.
            apply_positional_embedding (:obj:`bool`): A flag for applying positional
                embeddings.  Defaults to ``False``.
            positional_embedding_shape (pair of :obj:`int`): The height and width of
                the spatial positional embedding.
            positional_embedding_interpolation (:obj:`str`): The mode for the
                :class:`Interpolation` layer for resizing positional embeddings.
                Defaults to ``None``, which means no interpolation.
            blocks_per_regnet_stage (:obj:`int`): The number of blocks in each
                convolution stage.
            regnet_activation (:obj:`str`): The activation type for the convolution
                stages.
            norm_eps (:obj:`float`): The epsilon value for norm layers in the
                convolution stages.
            norm_bias (:obj:`bool`): A flag for including a bias parameter in the
                norm layers of convolution stages.  Defaults to ``False``.
            squeeze_excitation_reduced_dim (:obj:`int`): The "squeezed" dimension for
                :class:`.SqueezeExcitation` layers in the convolution stages.
            squeeze_excitation_bias (:obj:`bool`): A flag for including a bias parameter
                in the :class:`.SqueezeExcitation` layers.
            pooler_output_shape (pair of :obj:`int`): The height and width of the
                adaptive pooler's outputs.
            feed_forward_activation (:obj:`str`): The activation type for the
                feed-forward layer.
            num_prepend_tokens (:obj:`int`): The number of constant embedding tokens
                to prepend to the output sequence.
            num_append_tokens (:obj:`int`): The number of constant embedding tokens to
                append to the output sequence.

        Returns:
            The configured :obj:`.LayerBuilder`.
        """

        # pylint: disable=too-many-locals
        if convolution_output_dim is None:
            convolution_output_dim = input_dim
        if feed_forward_hidden_dim is None:
            feed_forward_hidden_dim = convolution_output_dim
        if output_dim is None:
            output_dim = input_dim

        kwargs = {}
        kwargs["drop_tokens"] = _basic.SelectByKey.Builder("convolution_tokens")
        kwargs["to_channels_first"] = _basic.ChannelsLastToFirst.Builder()

        if apply_positional_embedding:
            kwargs["positional_encoding"] = _SpatialPositionalEmbedding.Builder(
                positional_embedding_shape,
                dim=input_dim,
                interpolation=positional_embedding_interpolation,
            )

        regnet_kwargs = {
            "num_blocks": blocks_per_regnet_stage,
            "activation": regnet_activation,
            "norm": "layer_norm",
            "norm_eps": norm_eps,
            "norm_bias": norm_bias,
            "squeeze_excitation_reduced_dim": squeeze_excitation_reduced_dim,
            "squeeze_excitation_bias": squeeze_excitation_bias,
        }
        kwargs["convolution_0"] = _convolution.ResNetStage.create_regnet_builder(
            input_dim=input_dim,
            dim_per_group=1,
            output_dim=convolution_output_dim,
            **regnet_kwargs,
        )

        kwargs["pooler"] = SimpleAdaptiveAvgPooler.Builder(pooler_output_shape)

        kwargs["convolution_1"] = _convolution.ResNetStage.create_regnet_builder(
            input_dim=convolution_output_dim,
            output_dim=convolution_output_dim,
            dim_per_group=1,
            **regnet_kwargs,
        )

        kwargs["to_channels_last"] = _basic.ChannelsFirstToLast.Builder(
            flatten_spatial_dims=True
        )

        kwargs["feed_forward"] = _feed_forward.FeedForward.create_builder(
            input_dim=convolution_output_dim,
            hidden_dim=feed_forward_hidden_dim,
            output_dim=output_dim,
            activation=feed_forward_activation,
            hidden_transform_bias=True,
            output_transform_bias=True,
        )

        if num_prepend_tokens > 0 or num_append_tokens > 0:
            kwargs["add_tokens"] = _embedding.ConcatEmbedding.Builder(
                dim=output_dim,
                num_start=num_prepend_tokens,
                num_end=num_append_tokens,
            )

        return cls.Builder(**kwargs)


class AdaptiveConvPooler(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    An image pooling layer which performs a convolution and optional norm/activation, followed by a pooling layer.
    This layer expects inputs in NHWC format (channels-last).  The output has shape
    ``(batch size, sequence len, output dim)``, where the sequence length is determined by the pooler layer.
    """

    def __init__(
        self,
        *,
        to_channels_first: _ModuleOrBuilder,
        convolution: _ModuleOrBuilder,
        norm: _OptionalModuleOrBuilder = None,
        activation: _activation.ActivationSpecType = "relu",
        pooler: _ModuleOrBuilder,
        to_channels_last: _ModuleOrBuilder,
    ):
        activation = _activation.create_activation_layer(activation)
        layers = {
            "to_channels_first": to_channels_first,
            "convolution": convolution,
            "norm": norm,
            "activation": activation,
            "pooler": pooler,
            "to_channels_last": to_channels_last,
        }
        super().__init__(layers)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ,too-many-locals
        cls,
        *,
        input_dim: int,
        output_dim: _Optional[int] = None,
        kernel_size: _Union[int, _Tuple[int, int]],
        stride: _Union[int, _Tuple[int, int]] = 1,
        padding: _Union[str, int, _Tuple[int, int]] = 0,
        bias: bool = False,
        activation: _activation.ActivationSpecType = "relu",
        norm: str = "rms_norm",
        norm_bias: bool = False,
        norm_eps: float = 1e-5,
        pooler_output_shape: _Tuple[int, int],
    ):
        """
        Creates a builder for a :obj:`.AdaptiveConvPooler`.

        Args:
            input_dim (:obj:`int`): The number of features in the inputs.
            output_dim (:obj:`int`): The number of features in the outputs.  If ``None``, defaults to ``input_dim``.
            kernel_size (:obj:`Union[int, Tuple[int, int]]`): The kernel size of the convolution.
            stride (:obj:`Union[int, Tuple[int, int]]`): The stride of the convolution.  Defaults to ``1``.
            padding (:obj:`Union[str, int, Tuple[int, int]]`): The padding of the convolution.  Defaults to ``0``.
            bias (:obj:`bool`, optional): Whether to include a bias term in the convolution layer.  Defaults
                to ``False``.
            activation (:obj:`str`, optional): The activation type.  Defaults to ``"relu"``.
            norm (:obj:`str`, optional): The normalization type.  Defaults to ``"rms_norm"``.
            norm_bias (:obj:`bool`, optional): Whether to include a bias term in the normalization layer.  Defaults
                to ``False``.
            norm_eps (:obj:`float`, optional): The norm epsilon value.  Defaults to ``1e-5``.
            pooler_output_shape (pair of :obj:`int`): The height and width of the
                adaptive pooler's outputs.

        Returns:
            The configured :obj:`.LayerBuilder`.
        """
        if output_dim is None:
            output_dim = input_dim

        to_channels_first = _basic.ChannelsLastToFirst.Builder()
        convolution = _convolution.Conv2d.Builder(
            input_dim,
            output_dim,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
        )

        if norm is not None:
            norm = _norm.create_norm_builder(
                (output_dim,), norm, dim=1, bias=norm_bias, eps=norm_eps
            )

        pooler = SimpleAdaptiveAvgPooler.Builder(pooler_output_shape)

        to_channels_last = _basic.ChannelsFirstToLast.Builder(flatten_spatial_dims=True)

        return cls.Builder(
            to_channels_first=to_channels_first,
            convolution=convolution,
            norm=norm,
            activation=activation,
            pooler=pooler,
            to_channels_last=to_channels_last,
        )


class ConvPooler(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    An image pooling layer that applies the following layer sequence:

    * Channels last-to-first
    * Convolution
    * Optional padding
    * Norm
    * Activation
    * Average pooling 2D
    * Channels first-to-last
    """

    def __init__(
        self,
        *,
        to_channels_first: _ModuleOrBuilder,
        convolution: _ModuleOrBuilder,
        padding: _OptionalModuleOrBuilder,
        norm: _OptionalModuleOrBuilder = None,
        activation: _activation.ActivationSpecType = "relu",
        pooler: _ModuleOrBuilder,
        to_channels_last: _ModuleOrBuilder,
    ):
        activation = _activation.create_activation_layer(activation)
        layers = {
            "to_channels_first": to_channels_first,
            "convolution": convolution,
            "padding": padding,
            "norm": norm,
            "activation": activation,
            "pooler": pooler,
            "to_channels_last": to_channels_last,
        }
        super().__init__(layers)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ,too-many-locals
        cls,
        *,
        input_dim: int,
        output_dim: _Optional[int] = None,
        pad_to_multiple: _Optional[int] = None,
        convolution_kernel_size: _Union[int, _Tuple[int, int]],
        convolution_stride: _Union[int, _Tuple[int, int]] = 1,
        convolution_padding: _Union[str, int, _Tuple[int, int]] = 0,
        convolution_bias: bool = False,
        activation: _activation.ActivationSpecType = "relu",
        norm: str = "rms_norm",
        norm_bias: bool = False,
        norm_eps: float = 1e-5,
        pooler_kernel_size: _Union[int, _Tuple[int, int]],
        pooler_stride: _Union[int, _Tuple[int, int]] = 1,
    ):
        """
        Creates a builder for a :obj:`.ConvPooler`.

        Args:
            input_dim (:obj:`int`): The number of features in the inputs.
            output_dim (:obj:`int`): The number of features in the outputs.  If ``None``, defaults to ``input_dim``.
            pad_to_multiple (:obj:`int`, optional): An integer padding amount.  If specified, the layer pads
                the inputs (using top-left zero padding) so that the height and width are divisible by this value.
                This can be useful when the pooler has a nonzero stride.
            convolution_kernel_size (:obj:`Union[int, Tuple[int, int]]`): The kernel size of the convolution.
            convolution_stride (:obj:`Union[int, Tuple[int, int]]`): The stride of the convolution.  Defaults to
                ``1``.
            convolution_padding (:obj:`Union[str, int, Tuple[int, int]]`): The padding of the convolution.
                Defaults to ``0``.
            convolution_bias (:obj:`bool`, optional): Whether to include a bias term in the convolution layer.
                Defaults to ``False``.
            activation (:obj:`str`, optional): The activation type.  Defaults to ``"relu"``.
            norm (:obj:`str`, optional): The normalization type.  Defaults to ``"rms_norm"``.
            norm_bias (:obj:`bool`, optional): Whether to include a bias term in the normalization layer.  Defaults
                to ``False``.
            norm_eps (:obj:`float`, optional): The norm epsilon value.  Defaults to ``1e-5``.
            pooler_kernel_size (:obj:`Union[int, Tuple[int, int]]`): The kernel size of the average pooling layer.
            pooler_stride (:obj:`Union[int, Tuple[int, int]]`): The stride of the average pooling layer.  Defaults
                to ``1``.

        Returns:
            The configured :obj:`.LayerBuilder`.
        """
        if output_dim is None:
            output_dim = input_dim

        result = cls.Builder()

        result.to_channels_first = _basic.ChannelsLastToFirst.Builder()
        if pad_to_multiple:
            result.padding = _basic.PadToMultiple.Builder(
                multiple=pad_to_multiple, dim=(-2, -1)
            )
        result.convolution = _convolution.Conv2d.Builder(
            input_dim,
            output_dim,
            kernel_size=convolution_kernel_size,
            stride=convolution_stride,
            padding=convolution_padding,
            bias=convolution_bias,
        )
        if norm is not None:
            result.norm = _norm.create_norm_builder(
                (output_dim,), norm, dim=1, bias=norm_bias, eps=norm_eps
            )
        result.activation = _activation.create_activation_layer(activation)
        result.pooler = _torch_nn.AvgPool2d.Builder(
            kernel_size=pooler_kernel_size, stride=pooler_stride
        )
        result.to_channels_last = _basic.ChannelsFirstToLast.Builder(
            flatten_spatial_dims=True
        )

        return result



import torch as _torch

from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.mixture_of_experts.sampler import common as _common


class TopKMixtureOfExpertsSampler(_common.MixtureOfExpertsSampler, _LayerMixin):
    """
    A deterministic sampler that for each token selects the ``k`` experts with
    largest probability.  To compute the expert weights, the layer re-normalizes
    the probabilities of chosen experts so that they sum to ``1``.

    Args:
        num_experts_per_token (:obj:`int`): The number of experts to select
            for each token (i.e., ``k``).
    """

    def __init__(self, num_experts_per_token: int):
        super().__init__()
        self.num_experts_per_token = num_experts_per_token

    def extra_repr(self):
        return f"num_experts_per_token={self.num_experts_per_token}"

    def forward(
        self, probabilities: _torch.Tensor
    ) -> _common.MixtureOfExpertsSamplerOutput:
        weights, assignments = _torch.topk(
            probabilities, k=self.num_experts_per_token, dim=-1
        )
        weights = weights / weights.sum(dim=-1, keepdim=True)
        return _common.MixtureOfExpertsSamplerOutput(
            expert_assignments=assignments, assignment_weights=weights
        )



"""
This module implements the base converter type as well as some reusable components for
building converters.
"""

import abc as _abc
import collections as _collections
import functools as _functools
import re as _re
import warnings as _warnings
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm import _helpers
from tamm.typing import StateDictType as _StateDictType


class StateDictConverter(_abc.ABC):
    """
    An abstract interface for converting a |tamm| state dict to or from an
    "other" format.
    """

    @_abc.abstractmethod
    def is_tamm_key(self, key: str) -> bool:
        """Returns ``True`` if ``key`` is a |tamm| key and ``False`` otherwise."""

    @_abc.abstractmethod
    def is_other_key(self, key: str) -> bool:
        """
        Returns ``True`` if ``key`` is a key for the non-|tamm| format and ``False``
        otherwise.
        """

    def convert_from_tamm(self, state_dict: _StateDictType) -> _StateDictType:
        """Converts ``state_dict`` from |tamm| format to the other format."""
        unrecognized_keys = [key for key in state_dict if not self.is_tamm_key(key)]
        if len(unrecognized_keys) > 0:
            _warnings.warn(
                f"{self.__class__.__name__}.convert_from_tamm() called with "
                f"unrecognized keys: {unrecognized_keys}"
            )
        return self._convert_from_tamm_impl(state_dict)

    @_abc.abstractmethod
    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        """Converts ``state_dict`` from |tamm| format to the other format."""

    def convert_to_tamm(self, state_dict: _StateDictType) -> _StateDictType:
        """Converts ``state_dict`` to |tamm| format from the other format."""
        unrecognized_keys = [key for key in state_dict if not self.is_other_key(key)]
        if len(unrecognized_keys) > 0:
            _warnings.warn(
                f"{self.__class__.__name__}.convert_to_tamm() called with unrecognized "
                f"keys: {unrecognized_keys}"
            )
        return self._convert_to_tamm_impl(state_dict)

    @_abc.abstractmethod
    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        """Converts ``state_dict`` to tamm format from the other format."""


class UnionConverter(StateDictConverter):
    """
    A class for composing multiple independent :obj:`StateDictConverter` instances
    into a single converter.  Each converter takes responsibility for converting
    the subset of keys that match (according to :meth:`is_tamm_key` or
    :meth:`is_other_key`).

    If multiple converters output a converted weight with the same name, the merging
    mechanism takes the weight from the last converter in the converters list.

    Args:
        *converters: One or more converters of type :class:`StateDictConverter`.
    """

    def __init__(self, *converters):
        if len(converters) == 0:
            raise ValueError("UnionConverter requires at least one converter")
        self._converters = converters

    def is_tamm_key(self, key: str) -> bool:
        return any(c.is_tamm_key(key) for c in self._converters)

    def is_other_key(self, key: str) -> bool:
        return any(c.is_other_key(key) for c in self._converters)

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        sub_dicts = [
            {key: val for key, val in state_dict.items() if c.is_tamm_key(key)}
            for c in self._converters
        ]
        converted_sub_dicts = [
            c.convert_from_tamm(sub_dict)
            for c, sub_dict in zip(self._converters, sub_dicts)
        ]
        return _helpers.merge_dicts(*converted_sub_dicts)

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        sub_dicts = [
            {key: val for key, val in state_dict.items() if c.is_other_key(key)}
            for c in self._converters
        ]
        converted_sub_dicts = [
            c.convert_to_tamm(sub_dict)
            for c, sub_dict in zip(self._converters, sub_dicts)
        ]
        return _helpers.merge_dicts(*converted_sub_dicts)


class SequentialConverter(StateDictConverter):
    """
    A class for composing multiple independent :obj:`StateDictConverter` instances
    sequentially.  When converting from ``tamm``, the converter sequentially applies
    the ``from-tamm`` transformation of each subconverter, following the order of
    the subconverters during initialization of the parent converter.  When
    converting to ``tamm``, the converter applies the ``to-tamm`` transformation
    of each subconverter  in reverse order.

    Args:
        *converters: One or more converters of type :class:`StateDictConverter`.
    """

    def __init__(self, *converters):
        if len(converters) == 0:
            raise ValueError("SequentialConverter requires at least one converter")
        self._converters = converters

    def is_tamm_key(self, key: str) -> bool:
        return self._converters[0].is_tamm_key(key)

    def is_other_key(self, key: str) -> bool:
        return self._converters[-1].is_other_key(key)

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        for converter in self._converters:
            state_dict = converter.convert_from_tamm(state_dict)
        return state_dict

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        for converter in reversed(self._converters):
            state_dict = converter.convert_to_tamm(state_dict)
        return state_dict


class LayerwiseConverter(StateDictConverter):
    """
    A converter for state dicts with repeated layers.  For example, this class could
    help convert an "other" state dict with keys::

        [block0.weight0, block0.weight1, block1.weight0, block1.weight1]

    to a tamm state dict with keys::

        [layer_0.a.weight, layer_0.b.weight, layer_1.a.weight, layer_0.b.weight]

    Args:
        layer_converter (:obj:`StateDictConverter`): A converter for the repeated layer.
            For the example described above, this should convert an other state dict
            with keys ``["weight0", "weight1"]`` to one with keys
            ``["a.weight", "b.weight"]``.
        tamm_layer_prefix (:obj:`str`): The prefix for tamm layers, but with
            ``"<layer_idx>"`` substituted for the layer index.  This would be
            ``"layer_<layer_idx>."`` for the example above.
        other_layer_prefix (:obj:`str`): The prefix for "other" layers.  This would be
            ``"block<layer_idx>."`` for the example above.
        other_layer_idx_offset (:obj:`int`): An offset for the layer indexing of the
            other model.  Set to 1 if the other model uses 1-based indexing.
        max_other_layer_idx (:obj:`int`, optional): Matches until this layer index of
            the other model. Set to ``None`` if there's no matching limit.
        min_other_layer_idx (:obj:`int`, optional): Matches starting from this layer
            index of the other model.  Set to ``None`` to match from the smallest
            index.  Defaults to ``None``.
    """

    LAYER_IDX_PATTERN = "<layer_idx>"

    def __init__(
        self,
        layer_converter: StateDictConverter,
        tamm_layer_prefix: str,
        other_layer_prefix: str,
        other_layer_idx_offset: int = 0,
        max_other_layer_idx: _Optional[int] = None,
        min_other_layer_idx: _Optional[int] = None,
    ):
        self._layer_converter = layer_converter
        self._tamm_layer_prefix = tamm_layer_prefix
        self._other_layer_prefix = other_layer_prefix
        self._other_layer_idx_offset = other_layer_idx_offset
        self._max_other_layer_idx = max_other_layer_idx
        self._min_other_layer_idx = min_other_layer_idx

    @_functools.cached_property
    def _tamm_layer_prefix_regex(self):
        return self._get_regex_from_layer_prefix(self._tamm_layer_prefix)

    @_functools.cached_property
    def _other_layer_prefix_regex(self):
        return self._get_regex_from_layer_prefix(self._other_layer_prefix)

    @classmethod
    def _get_regex_from_layer_prefix(cls, layer_prefix):
        escaped_prefix = _re.escape(layer_prefix)
        escaped_layer_idx_pattern = _re.escape(cls.LAYER_IDX_PATTERN)
        pattern = escaped_prefix.replace(escaped_layer_idx_pattern, r"([\d]+)")
        return _re.compile(pattern)

    def is_tamm_key(self, key: str) -> bool:
        match = self._match_prefix(self._tamm_layer_prefix_regex, key)
        if match is None:
            return False
        _, remaining_key = match
        return self._layer_converter.is_tamm_key(remaining_key)

    def is_other_key(self, key: str) -> bool:
        match = self._match_prefix(self._other_layer_prefix_regex, key)
        if match is None:
            return False
        _, remaining_key = match
        return self._layer_converter.is_other_key(remaining_key)

    @staticmethod
    def _match_prefix(compiled_prefix_regex, key):
        """
        Returns (layer_idx, remaining_key) if key matches the regex and None otherwise.
        Here remaining_key is everything that comes after the prefix.
        """
        match = compiled_prefix_regex.match(key)
        if match is None:
            return None
        layer_idx = match.group(1)
        remaining_key = key[match.end() :]
        return layer_idx, remaining_key

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        state_subdicts_by_layer = self._get_state_subdicts_by_layer(
            self._other_layer_prefix_regex, state_dict
        )
        result = {}
        for layer_idx, state_subdict in state_subdicts_by_layer.items():
            if self._min_other_layer_idx is not None:
                if int(layer_idx) < self._min_other_layer_idx:
                    continue
                layer_idx = str(
                    int(layer_idx)
                    - self._other_layer_idx_offset
                    - self._min_other_layer_idx
                )
            else:
                layer_idx = str(int(layer_idx) - self._other_layer_idx_offset)
            converted_state = self._layer_converter.convert_to_tamm(state_subdict)
            prefix = self._tamm_layer_prefix.replace(self.LAYER_IDX_PATTERN, layer_idx)
            converted_state = {prefix + k: v for k, v in converted_state.items()}
            result.update(converted_state)
        return result

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        state_subdicts_by_layer = self._get_state_subdicts_by_layer(
            self._tamm_layer_prefix_regex, state_dict
        )
        result = {}
        for layer_idx, state_subdict in state_subdicts_by_layer.items():
            if self._min_other_layer_idx is not None:
                layer_idx = str(
                    int(layer_idx)
                    + self._other_layer_idx_offset
                    + self._min_other_layer_idx
                )
            else:
                layer_idx = str(int(layer_idx) + self._other_layer_idx_offset)
            converted_state = self._layer_converter.convert_from_tamm(state_subdict)
            prefix = self._other_layer_prefix.replace(self.LAYER_IDX_PATTERN, layer_idx)
            converted_state = {prefix + k: v for k, v in converted_state.items()}
            result.update(converted_state)
        return result

    def _get_state_subdicts_by_layer(self, compiled_prefix_regex, state_dict):
        """
        Returns a dictionary that maps the layer index for each layer to the state dict
        for that layer.  The keys in each state dict no longer contain the prefix.
        """
        result = _collections.defaultdict(dict)
        for key, param in state_dict.items():
            match = self._match_prefix(compiled_prefix_regex, key)
            if match is None:
                raise ValueError(f"{key} did not match {compiled_prefix_regex}")
            layer_idx, remaining_key = match
            if (
                self._max_other_layer_idx is not None
                and int(layer_idx) > self._max_other_layer_idx
            ):
                continue
            result[layer_idx][remaining_key] = param
        return result


class ParamMapper:
    """
    Defines a reversible mapping of a parameter.  The ``to_tamm_fn``
    and ``from_tamm_fn`` functions should be inverses of one another.

    Args:
        to_tamm_fn: The function for mapping the param from "other" to |tamm| format.
        from_tamm_fn: The function for mapping the param from |tamm| to "other" format.
    """

    def __init__(
        self,
        to_tamm_fn: _Callable[
            [_torch.Tensor], _Union[_torch.Tensor, _List[_torch.Tensor]]
        ],
        from_tamm_fn: _Callable[
            [_torch.Tensor], _Union[_torch.Tensor, _List[_torch.Tensor]]
        ],
    ):
        self._to_tamm_fn = to_tamm_fn
        self._from_tamm_fn = from_tamm_fn

    def to_tamm(self, *args):
        """Map the parameter to |tamm| format."""
        return self._to_tamm_fn(*args)

    def from_tamm(self, *args):
        """Map the parameter from |tamm| format."""
        return self._from_tamm_fn(*args)


class StringMatchingConverter(StateDictConverter):
    """
    Converts state dicts by mapping keys from one string to another
    optionally transforming the params.

    This class manages multiple sets of one-to-one parameter mappings. For
    more complex scenarios, consider using
    :class:`ManyToManyStringMatchingConverter`.

    Args:
        tamm_to_other_keys (:obj:`dict`): A dictionary that maps |tamm| keys to
            other keys.
        param_mappers (:obj:`dict`): An optional dictionary that maps a subset of
            tamm keys to :obj:`ParamMapper` objects for transforming weights.
    """

    def __init__(
        self,
        tamm_to_other_keys: _Dict[str, str],
        *,
        param_mappers: _Dict[str, "ParamMapper"] = None,
    ):
        self._tamm_to_other_keys = tamm_to_other_keys
        self._param_mappers = {} if param_mappers is None else param_mappers

    @property
    def _tamm_keys(self):
        return self._tamm_to_other_keys.keys()

    @property
    def _other_keys(self):
        return self._tamm_to_other_keys.values()

    def is_tamm_key(self, key: str) -> bool:
        return key in self._tamm_keys

    def is_other_key(self, key: str) -> bool:
        return key in self._other_keys

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        tamm_state_dict = {}
        for other_key, other_param in state_dict.items():
            tamm_keys = [
                k for k, v in self._tamm_to_other_keys.items() if v == other_key
            ]
            for tamm_key in tamm_keys:
                mapper = self._param_mappers.get(tamm_key)
                tamm_param = other_param
                if mapper is not None:
                    tamm_param = mapper.to_tamm(other_param)
                tamm_state_dict[tamm_key] = tamm_param
        return tamm_state_dict

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        other_state_dict = {}
        for tamm_key, tamm_param in state_dict.items():
            other_key = self._tamm_to_other_keys[tamm_key]

            mapper = self._param_mappers.get(tamm_key)
            other_param = tamm_param
            if mapper is not None:
                other_param = mapper.from_tamm(tamm_param)
            other_state_dict[other_key] = other_param
        return other_state_dict


class ManyToManyStringMatchingConverter(StateDictConverter):
    """
    This class serves as an addition to :class:`StringMatchingConverter` to handle
    one-to-many, many-to-one, or many-to-many parameter relationships between |tamm|
    and other formats.

    Note that this class only manages one type of mapping relationship at a time,
    whereas :class:`StringMatchingConverter` handles multiple mappings
    simultaneously.

    Args:
        tamm_keys (:obj:`list` of :obj:`str`): A list of |tamm| keys to map.
        other_keys (:obj:`list` of :obj:`str`): A list of keys in other formats
            to map.
        param_mapper: A :class:`ParamMapper` that accepts the |tamm| weights
            and outputs the other weights (and vice versa).

    Example:

        Use ``fused_qkv_converter`` to combine separate query, key, and value
        weights into a single ``fused_qkv`` weight.

        .. code-block:: python

            param_mapper = ParamMapper(
                to_tamm_fn=lambda *tensors: torch.stack(tensors),
                from_tamm_fn=torch.unbind,
            )

            fused_qkv_converter = ManyToManyStringMatchingConverter(
                ["fused_qkv"], ["Q", "K", "V"], param_mapper
            )
    """

    def __init__(
        self,
        tamm_keys: _List[str],
        other_keys: _List[str],
        param_mapper: _Optional[ParamMapper] = None,
    ):
        self._tamm_keys = tamm_keys
        self._other_keys = other_keys
        self._param_mapper = param_mapper
        if len(self._tamm_keys) != len(self._other_keys) and self._param_mapper is None:
            raise ValueError(
                f"Lengths of tamm_keys ({len(self._tamm_keys)}) and "
                f"other_keys {len(self._other_keys)}) do not match. Please provide a "
                f"mapper when they are not of equal length."
            )

    def is_tamm_key(self, key: str) -> bool:
        return key in self._tamm_keys

    def is_other_key(self, key: str) -> bool:
        return key in self._other_keys

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        other_params = [state_dict[key] for key in self._other_keys]
        if self._param_mapper is not None:
            tamm_params = self._param_mapper.to_tamm(*other_params)
            if not isinstance(tamm_params, (list, tuple)):
                tamm_params = [tamm_params]
        else:
            tamm_params = other_params
        if len(self._tamm_keys) != len(tamm_params):
            raise ValueError(
                f"Lengths of tamm_keys ({len(self._tamm_keys)}) and the converted "
                f"tamm_params ({len(tamm_params)}) do not match."
            )
        return dict(zip(self._tamm_keys, tamm_params))

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        tamm_params = [state_dict[key] for key in self._tamm_keys]
        if self._param_mapper is not None:
            other_params = self._param_mapper.from_tamm(*tamm_params)
            if not isinstance(other_params, (list, tuple)):
                other_params = [other_params]
        else:
            other_params = tamm_params
        if len(self._other_keys) != len(other_params):
            raise ValueError(
                f"Lengths of other_keys ({len(self._other_keys)}) and the converted "
                f"other_params ({len(other_params)}) do not match."
            )
        return dict(zip(self._other_keys, other_params))


class PrefixStringMatchingConverter(StateDictConverter):
    """
    A converter wrapper that handles prefixes in front of keys.  When converting
    to |tamm|, this converter:

    1. Creates a sub dictionary of all weights with names starting with
       ``other_prefix``.
    2. Replaces the sub dictionary keys with only their suffixes.
    3. Converts the sub dictionary with ``converter``.
    4. Prepends ``tamm_prefixes`` to the keys of the result.

    Args:
        converter (:obj:`StateDictConverter`): A converter for the sub
            dictionary.
        tamm_prefix (:obj:`str`): Prefix for the |tamm| keys.
        other_prefix (:obj:`str`): Prefix for the other keys.
    """

    def __init__(
        self, *, converter: StateDictConverter, tamm_prefix: str, other_prefix: str
    ):
        self._converter = converter
        self._tamm_prefix = tamm_prefix
        self._other_prefix = other_prefix

    def is_tamm_key(self, key: str) -> bool:
        if not key.startswith(self._tamm_prefix):
            return False
        suffix = key[len(self._tamm_prefix) :]
        return self._converter.is_tamm_key(suffix)

    def is_other_key(self, key: str) -> bool:
        if not key.startswith(self._other_prefix):
            return False
        suffix = key[len(self._other_prefix) :]
        return self._converter.is_other_key(suffix)

    def _convert_from_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        state_dict = {
            k: v for k, v in state_dict.items() if k.startswith(self._tamm_prefix)
        }
        state_dict = _helpers.strip_prefix_from_state_dict_keys(
            state_dict, self._tamm_prefix
        )
        converted_dict = self._converter.convert_from_tamm(state_dict)
        converted_dict = _helpers.add_prefix_to_state_dict_keys(
            converted_dict, self._other_prefix
        )
        return converted_dict

    def _convert_to_tamm_impl(self, state_dict: _StateDictType) -> _StateDictType:
        state_dict = {
            k: v for k, v in state_dict.items() if k.startswith(self._other_prefix)
        }
        state_dict = _helpers.strip_prefix_from_state_dict_keys(
            state_dict, self._other_prefix
        )
        converted_dict = self._converter.convert_to_tamm(state_dict)
        converted_dict = _helpers.add_prefix_to_state_dict_keys(
            converted_dict, self._tamm_prefix
        )
        return converted_dict


class MultiPrefixStringMatchingConverter(UnionConverter):
    """
    A composition of :class:`PrefixStringMatchingConverter` converters
    to support multiple pairs of prefixes.  For example, this class could help
    convert an "other" state dict with keys::

        [block0.0.attention, block0.1.attention, block1.0.attention]

    to a |tamm| state dict with keys::

        [layer_0.attention, layer_1.attention, layer_2.attention]

    For this case, ``converter`` converts attention layers, while
    ``tamm_to_other_prefixes`` is the dictionary
    ``{"layer_0": "block0.0", "layer_1": "block0.1", "layer_2": "block1.0"}``.

    Args:
        converter (:obj:`StateDictConverter`): A converter for the selected
            weights.
        tamm_to_other_prefixes (:obj:`dict`): A dictionary that maps |tamm|
            prefixes to other prefixes.
    """

    def __init__(
        self, *, converter: StateDictConverter, tamm_to_other_prefixes: _Dict[str, str]
    ):
        converters = [
            PrefixStringMatchingConverter(
                tamm_prefix=tamm_prefix, other_prefix=other_prefix, converter=converter
            )
            for tamm_prefix, other_prefix in tamm_to_other_prefixes.items()
        ]
        super().__init__(*converters)



"""
Implements a residual connection layers for :class:`tamm.layers.Sequential`.
"""

from typing import Optional as _Optional

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm.layers import activation as _activation
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class ResidualAdd(_nn.Module, _LayerMixin):
    """
    A simple module for implementing residual connections.  The forward function takes
    ``input`` and ``residual_input`` arguments and returns their sum.
    """

    def __init__(self):
        super().__init__()  # Needed because builder doesn't support *args

    def forward(
        self,
        input: _torch.Tensor,  # pylint: disable=redefined-builtin
        residual_input: _torch.Tensor,
    ) -> _torch.Tensor:
        return input + residual_input


class ShortcutAddActResidualConnection(_nn.Module, _LayerMixin):
    """
    A residual connection that applies a shortcut transformation to the residual
    inputs, then adds the shortcut result to outputs, and finally applies an activation
    transform to the sum.
    """

    def __init__(
        self,
        shortcut: _OptionalModuleOrBuilder = None,
        activation: _Optional[_activation.ActivationSpecType] = None,
    ):
        super().__init__()
        self.shortcut = _helpers.maybe_build_module(shortcut)
        self.activation = _activation.create_activation_layer(activation)

    # pylint: disable=redefined-builtin
    def forward(
        self,
        input: _torch.Tensor,
        residual_input: _torch.Tensor,
    ) -> _torch.Tensor:
        if self.shortcut is not None:
            residual_input = self.shortcut(residual_input)
        x = input + residual_input
        if self.activation is None:
            return x
        return self.activation(x)


class GatedActivationResidualConnection(_nn.Module, _LayerMixin):
    """
    A residual connection that applies gated activation to the residual inputs (using
    the non-residual inputs for gating).
    """

    def __init__(self, activation: _activation.ActivationSpecType):
        super().__init__()
        self.activation = _activation.create_activation_layer(activation)

    # pylint: disable=redefined-builtin
    def forward(
        self,
        input: _torch.Tensor,
        residual_input: _torch.Tensor,
    ) -> _torch.Tensor:
        return self.activation(gate_input=input, input=residual_input)



"""
utils.transformers_utils
========================

This module implements utilities for working with Hugging Face checkpoints.

.. autofunction:: tamm.utils.transformers_utils.load_transformers_state_dict

.. autofunction:: tamm.utils.transformers_utils.save_transformers_state_dict

"""

import glob as _glob
import logging as _logging
import os as _os
import shutil as _shutil
import tempfile as _tempfile
from typing import Generator as _Generator

import torch as _torch

from tamm import _helpers
from tamm.typing import OptionalDeviceOrString as _OptionalDeviceOrString
from tamm.typing import PathLike as _PathLike
from tamm.typing import StateDictType as _StateDictType
from tamm.utils import user_dir_utils as _user_dir_utils

_logger = _logging.getLogger(__name__)


def load_transformers_state_dict(
    checkpoint_dir: _PathLike, *, device: _OptionalDeviceOrString = None
) -> _StateDictType:
    """
    Loads model state saved by :meth:`PreTrainedModel.save_pretrained` from
    ``transformers``.  This only loads the model's state dict, not its config.
    It supports ``torch`` native and ``safetensors`` formats as well as sharded
    checkpoints.

    Args:
        checkpoint_dir (:obj:`PathLike`): The local checkpoint directory containing
            the serialized model state.
        device (:obj:`str` or :obj:`torch.device`, optional): The target device for the
            loaded tensors.

    Returns:
        A :obj:`dict` that maps tensor names to loaded tensors.
    """
    state_dicts = list(_load_transformers_state_dicts(checkpoint_dir, device=device))
    return _helpers.merge_dicts(*state_dicts)


def _load_transformers_state_dicts(
    checkpoint_dir: _PathLike, *, device: _OptionalDeviceOrString = None
) -> _Generator[_StateDictType, None, None]:
    """
    Generates a sequence of loaded state dicts from ``checkpoint_dir``, one for each
    shard in the checkpoint.
    """
    checkpoint_dir = _helpers.get_expanded_abspath(checkpoint_dir)
    for path in _get_native_torch_model_paths(checkpoint_dir):
        yield _load_native_torch(path, device=device)
    for path in _get_safetensors_model_paths(checkpoint_dir):
        yield _load_safetensors(path, device=device)


def _load_native_torch(
    path: _PathLike, *, device: _OptionalDeviceOrString = None
) -> _StateDictType:
    _logger.info("Loading %s", path)
    return _helpers.safe_torch_load(path, map_location=device)


def _load_safetensors(
    path: _PathLike, *, device: _OptionalDeviceOrString = None
) -> _StateDictType:
    import safetensors.torch  # pylint: disable=import-outside-toplevel

    _logger.info("Loading %s", path)
    if device is not None:
        device = _torch.device(device).type  # safetensors errors out if not for this
    return safetensors.torch.load_file(path, device=device)


def _get_native_torch_model_paths(checkpoint_dir):
    pattern = _os.path.join(checkpoint_dir, "pytorch_model*.bin")
    return _glob.glob(pattern)


def _get_safetensors_model_paths(checkpoint_dir):
    pattern = _os.path.join(checkpoint_dir, "model*.safetensors")
    return _glob.glob(pattern)


def _get_all_model_paths(checkpoint_dir):
    paths1 = _get_native_torch_model_paths(checkpoint_dir)
    paths2 = _get_safetensors_model_paths(checkpoint_dir)

    pattern = _os.path.join(checkpoint_dir, "*model*.index.json")
    paths3 = _glob.glob(pattern)

    return paths1 + paths2 + paths3


def save_transformers_state_dict(
    state_dict: _StateDictType,
    checkpoint_dir: _PathLike,
    max_shard_size="5GB",
    **kwargs,
) -> None:
    """
    Saves a model's state dict in Hugging Face format.  This only saves the state dict,
    not the model's config, tokenizer, etc.

    Args:
        state_dict (:obj:`dict`): A state dictionary returned by
            :meth:`nn.Module.state_dict`.
        checkpoint_dir (:obj:`PathLike`): The target directory for saving the state.
        max_shard_size (:obj:`str`, optional): The ``max_shard_size`` argument for
            :meth:`PreTrainedModel.save_pretrained` from ``transformers``.  Defaults to
            ``"5GB"``.
        **kwargs: Additional keyword arguments for
            :meth:`PreTrainedModel.save_pretrained`.
    """
    checkpoint_dir = _helpers.get_expanded_abspath(checkpoint_dir)
    _logger.info("Saving to checkpoint to %s", checkpoint_dir)
    with _tempfile.TemporaryDirectory(
        dir=_user_dir_utils.get_tamm_tmp_dir()
    ) as temp_dir:
        dummy_model = _create_dummy_hf_model()
        dummy_model.save_pretrained(
            save_directory=temp_dir,
            state_dict=state_dict,
            max_shard_size=max_shard_size,
            **kwargs,
        )

        _os.makedirs(checkpoint_dir, exist_ok=True)
        for path in _get_all_model_paths(temp_dir):
            basename = _os.path.basename(path)
            dest = _os.path.join(checkpoint_dir, basename)
            _shutil.move(path, dest)


def _create_dummy_hf_model():
    """Creates and returns a small hf transformers model."""
    import transformers  # pylint: disable=import-outside-toplevel

    config = transformers.BertConfig(
        vocab_size=1,
        hidden_size=1,
        num_hidden_layers=1,
        num_attention_heads=1,
        intermediate_size=1,
        max_position_embeddings=1,
    )
    return transformers.BertModel(config)



"""
tamm.preprocessors
------------------

This module implements configurable preprocessors that
process raw data into model inputs.

.. Tip::

    Users new to this module should also check out our
    :ref:`preprocessor tutorial <preprocessors_guide>`.

.. autoclass:: tamm.preprocessors.Preprocessor
    :members:
    :special-members: __call__
    :private-members:

.. autofunction:: tamm.preprocessors.create
.. autofunction:: tamm.preprocessors.list
.. automodule:: tamm.preprocessors.registry

"""
from tamm.preprocessors import instruct_lm

# pylint: disable=redefined-builtin
from tamm.preprocessors.api import create, describe
from tamm.preprocessors.api import list_objects as list
from tamm.preprocessors.base import Preprocessor

__all__ = [
    "Preprocessor",
    "list",
    "create",
    "describe",
    "instruct_lm",
]



import logging as _logging
from typing import Union as _Union

from torch import nn as _nn

from tamm._adapters_v1 import ModelAdapter as _ModelAdapterV1
from tamm._adapters_v1.adapter_api import get_all_adapter_ids as _get_all_adapter_ids
from tamm._adapters_v1.adapter_api import set_active_adapter as _set_active_adapter
from tamm.layers.common.post_hooks.common import CompositePostHook, IdentityPostHook

_logger = _logging.getLogger(__name__)


class ModelAdapterPostHook:
    def __init__(
        self,
        model_adapter: "_ModelAdapterV1",
        *,
        adapter_id: str,
    ):
        self.model_adapter = model_adapter
        self.adapter_id = adapter_id

    def __call__(self, model: _nn.Module) -> _nn.Module:
        self.model_adapter.adapt_model(
            model,
            adapter_id=self.adapter_id,
        )
        return model


class ActivateAdapterPostHook:
    def __init__(
        self,
        active_adapter: _Union[int, str] = 0,
    ):
        self.active_adapter = active_adapter

    def __call__(self, model: _nn.Module) -> _nn.Module:
        if isinstance(self.active_adapter, int):
            model_adapters = _get_all_adapter_ids(model)
            if self.active_adapter >= len(model_adapters):
                raise IndexError(
                    "Not enough adapters present in the model. "
                    f"Trying to activate adapter with index {self.active_adapter} "
                    f"while only {len(model_adapters)} adapters "
                    "present in the model."
                )
            self.active_adapter = model_adapters[self.active_adapter]
        _set_active_adapter(model, self.active_adapter)
        return model


def get_model_adapters_post_hook(
    adapters: dict,
    active_adapter: _Union[int, str] = 0,
):
    if adapters is None:
        return IdentityPostHook()

    hooks = [
        ModelAdapterPostHook(model_adapter, adapter_id=adapter_id)
        for adapter_id, model_adapter in adapters.items()
    ]
    hooks.append(ActivateAdapterPostHook(active_adapter=active_adapter))

    return CompositePostHook(*hooks)



import dataclasses as _dataclasses
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class OutputWithSideOutputs:
    """
    Class to support passing side outputs to higher level.
    This is used in layers in tamm.layers.sequential.Sequential to capture
    intermediate outputs that might be helpful for users in training process.
    """

    output: _torch.Tensor
    side_outputs: _Optional[_Dict[str, _Any]] = _dataclasses.field(default_factory=dict)

    def merge_side_outputs(self, other_side_outputs: _Dict[str, _Any]) -> None:
        self.side_outputs = merge_side_outputs(self.side_outputs, other_side_outputs)


def merge_side_outputs(
    side_outputs: _Dict[str, _Any], other_side_outputs: _Dict[str, _Any]
):
    result = {**side_outputs}

    for key, other_val in other_side_outputs.items():
        if key not in side_outputs:
            result[key] = other_val
            continue

        val = side_outputs[key]
        if not isinstance(val, list):
            val = [val]
        if isinstance(other_val, list):
            val.extend(other_val)
        else:
            val.append(other_val)
        result[key] = val

    return result



import abc
import json
import logging as _logging
from typing import Any, Dict, Protocol

from tamm.runtime_configuration import rc as _rc

_logger = _logging.getLogger(__name__)

_USER_DEFINED_NORMALIZERS = []


class SupportsObjectHook(Protocol):
    """
    A protocol for objects that apply normalization before
    :func:`tamm.utils.json.loads` or denormalization after
    :func:`tamm.utils.json.dumps`.
    """

    @abc.abstractmethod
    def object_hook(self, json_obj: Any) -> Any:
        """
        A hook for modifying a JSON-compatible object (a collection of
        ``dict``, ``list``, ``str``, ``int``, ``float``, ``bool``, and
        ``None``).

        Args:
            json_obj: A JSON-compatible object.

        Returns:
            A possibly modified version of the input ``json_obj`` that
            is still JSON-compatible.
        """


def register_load_normalizer(normalizer: SupportsObjectHook) -> None:
    """
    A function that modifies :func:`tamm.utils.json.loads` by registering a
    normalizer.  The ``loads()`` function will map every JSON object with the
    normalizer's ``object_hook()`` method prior to |tamm| deserialization.

    Args:
        normalizer (:obj:`SupportsObjectHook`): The normalizer to register.
    """
    _USER_DEFINED_NORMALIZERS.append(normalizer)


class BackwardCompatibilityTypeNormalizer(SupportsObjectHook):
    """
    A normalizer for supporting backward compatibility of loading |tamm|
    types that have been renamed or relocated.  The normalizer updates
    ``__tamm_type__`` prefixes based on string matching.

    Args:
        old_to_new_types (:obj:`dict`): A :obj:`dict` that maps old
            ``__tamm_type__`` prefixes to new ones.  The
            :meth:`.object_hook` method applies string substitution
            to ``json_obj["__tamm_type__"]`` if this string exists and
            starts with a key in ``old_to_new_types``.  The substitution
            process executes in the order of the ``old_to_new_types``
            entries, and it may result in multiple substitutions.
    """

    def __init__(self, old_to_new_types: Dict[str, str]):
        self._old_to_new_types = old_to_new_types

    def object_hook(self, json_obj: Any) -> Any:
        if not isinstance(json_obj, dict) or "__tamm_type__" not in json_obj:
            return json_obj

        tamm_type = json_obj["__tamm_type__"]

        if not isinstance(tamm_type, str):
            _logger.debug("json_obj contains a non-str __tamm_type__: %s", json_obj)
            return json_obj

        for old_type, new_type in self._old_to_new_types.items():
            if tamm_type.startswith(old_type):
                tamm_type = tamm_type.replace(old_type, new_type, 1)
                json_obj["__tamm_type__"] = tamm_type
                _logger.debug(
                    "Migrated %s type to %s for deserializing JSON obj: %s",
                    old_type,
                    new_type,
                    json_obj,
                )

        return json_obj


class TorchDtypeNormalizer(SupportsObjectHook):
    """
    Our serialization scheme for dtypes is robust, but it is unnatural for
    users editing dtype fields "by hand".  The natural thing is to add
    "dtype": "bfloat16", but so far our scheme above requires
    "dtype": {"__tamm_type__": "torch.dtype", "name": "bfloat16"}.
    This is a best-effort attempt to also support the natural way.
    """

    def object_hook(self, json_obj: Any) -> Any:
        """
        Scans through obj.items().  If the key ends with "dtype" and the value is a
        string dtype (such as "torch.float16" or "float16"), then we convert the value
        to the corresponding :obj:`torch.dtype`.

        Args:
            json_obj: Input JSON object.

        Returns: normalized obj if obj is recognized by normalizer

        """
        if not isinstance(json_obj, dict):
            return json_obj
        for key, val in json_obj.items():
            if not isinstance(key, str):
                continue
            if not isinstance(val, str):
                continue
            if not (key.startswith("dtype") or key.endswith("dtype")):
                continue
            json_obj[key] = {
                "__tamm_type__": "torch.dtype",
                "name": val[6:] if val.startswith("torch.") else val,
            }
        return json_obj


class CompositeJSON2JSONConverter:
    def __init__(self, *object_hooks: "SupportsObjectHook"):
        """
        Each ``object_hook`` processes a JSON object into another schema
        also in JSON. ``object_hook`` is responsible for filtering and transforming
        JSON objects.

        For example, the following object hook converts
        all string values in JSON to UPPER case:

        .. code-block:: python

        class UppercaseNormalizer(SupportsObjectHook):
            def object_hook(self, json_obj: dict):
                 if not isinstance(json_obj, str):
                    return json_obj
                 return json_obj.upper()

        Args:
            *object_hooks: instance of class that provides object_hook()
        """
        self._object_hooks = object_hooks

    def __call__(
        self,
        json_str: str,
        *,
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        cls=None,
        indent=None,
        separators=None,
        default=None,
        sort_keys=False,
        **kw,
    ) -> str:
        """
        Implements string-based JSON-JSON conversion by exploiting the
        ``object_hook`` mechanism provided by Python json standard library.
        A json string is loaded with object hooks applied in sequence and dumped again
        for subsequent uses.
        """
        for n in self._object_hooks:
            json_str = json.dumps(
                json.loads(json_str, object_hook=n.object_hook),
                skipkeys=skipkeys,
                ensure_ascii=ensure_ascii,
                check_circular=check_circular,
                allow_nan=allow_nan,
                cls=cls,
                indent=indent,
                default=default,
                separators=separators,
                sort_keys=sort_keys,
                **kw,
            )
        return json_str


class ModulePrefixNormalizer(SupportsObjectHook):
    @classmethod
    def _get_prefixed_namespace(cls, namespace: str, delimiter=":"):
        return _rc.PROJECT_SLUG + delimiter + namespace

    def object_hook(self, json_obj: Any) -> Any:
        """
        Prepends the |tamm| module prefix to ``__tamm_type__`` object key for tamm-based
        serializable objects if not already prefixed to ensure backward compatibility
        with current |tamm| JSON registry schema notation
        ``tamm:<json_namespace>`:<registry_class>``.

        Args:
            json_obj: Input JSON object.

        Returns:
            Output dictionary with ``__tamm_type__`` values maybe prefixed with ``tamm``
                module prefix.
        """
        # pylint: disable=duplicate-code
        if not isinstance(json_obj, dict):
            return json_obj
        tamm_type_info = json_obj.get("__tamm_type__", None)
        if not tamm_type_info:
            return json_obj

        _delimiter = ":"

        if _delimiter not in tamm_type_info:
            # skip __tamm_type__ which does not contain colon (:),
            # e.g. {'__tamm_type__': 'torch.dtype', 'name': 'float16'}
            return json_obj

        # __tamm_type__ is typically of the format:
        # <module_prefix>:<json_namespace>:<serializable_registry_class>
        _, _, *args = tamm_type_info.split(_delimiter)

        if not args:
            json_obj["__tamm_type__"] = ModulePrefixNormalizer._get_prefixed_namespace(
                namespace=json_obj["__tamm_type__"], delimiter=_delimiter
            )
        return json_obj


class ModulePrefixDeNormalizer(SupportsObjectHook):
    def object_hook(self, json_obj: Any) -> Any:
        """
        Strip off tamm: prefix for backward compatibility
        """
        if not isinstance(json_obj, dict):
            return json_obj
        tamm_type_info = json_obj.get("__tamm_type__", None)
        if not tamm_type_info:
            return json_obj
        tamm_type_info = str(tamm_type_info)
        if tamm_type_info.startswith(f"{_rc.PROJECT_SLUG}:"):
            json_obj["__tamm_type__"] = tamm_type_info[len(f"{_rc.PROJECT_SLUG}:") :]
        return json_obj


def get_normalizer():
    """
    Get the JSON-JSON converter which can be applied before de-serialization.
    Typical purpose of this normalization is to convert known legacy JSON schema
    generated by previous tamm releases to something compatible with this version.
    """

    normalizers = [
        ModulePrefixNormalizer(),
        TorchDtypeNormalizer(),
    ]
    return CompositeJSON2JSONConverter(*normalizers, *_USER_DEFINED_NORMALIZERS)


def get_denormalizer():
    """
    Get the JSON-JSON converter which can be applied after serialization. Typical
    purpose of this denormalization is to remove incompatible JSON schema for backward
    compatibility with older tamm releases.
    """
    return CompositeJSON2JSONConverter(ModulePrefixDeNormalizer())



"""
layers.positional_encoding
--------------------------

This module implements layers for encoding positions into tensors.

.. autoclass:: tamm.layers.AbsolutePositionalEmbedding
    :show-inheritance:
    :members: create_builder, forward

.. autoclass:: tamm.layers.SpatialPositionalEmbedding
    :show-inheritance:
    :members: forward
"""

import enum as _enum
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

import tamm.layers.functional
from tamm import _helpers
from tamm.layers import basic as _basic
from tamm.layers import embedding as _embedding
from tamm.layers import init as _init
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import ModuleBuilder as _ModuleBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class AbsolutePositionalEmbedding(_torch.nn.Module, _ConfigurableLayerMixin):
    """
    Layer for absolute positional embeddings.  This layer maps positions to trainable
    embedding tensors and then adds these embeddings to the token embeddings.

    Args:
        embedding: An embedding layer (or builder) that maps positions to embeddings.
        sequence_start (:obj:`int`, optional): An option that controls which tokens
            receive positional embeddings.  If ``sequence_start > 0``, then the first
            ``sequence_start`` tokens do not receive positional embeddings.
    """

    def __init__(
        self,
        embedding: _ModuleBuilder,
        sequence_start: _Optional[int] = None,
    ):
        super().__init__()
        self.embedding = _helpers.maybe_build_module(embedding)
        self.sequence_start = sequence_start

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        num_embeddings: int,
        embedding_dim: int,
        *,
        sequence_start: _Optional[int] = None,
    ) -> _ModuleBuilder:
        """
        Creates and returns a builder for :class:`AbsolutePositionalEmbedding` layers.

        Args:
            num_embeddings (:obj:`int`): The number of embeddings in the positional
                embedding table.
            embedding_dim (:obj:`int`): The dimension of the embeddings.
            sequence_start (:obj:`int`, optional): The ``sequence_start`` option for the
                layer.

        Returns:
            The configured :obj:`LayerBuilder`.
        """
        embedding = _embedding.Embedding.Builder(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        return cls.Builder(
            embedding=embedding,
            sequence_start=sequence_start,
        )

    def _maybe_broadcast_embedding(
        self, inputs: _torch.Tensor, pos_embeddings: _torch.Tensor
    ) -> _torch.Tensor:
        """
        If there's a dimension mismatch between inputs and pos_embeddings,
        we broadcast pos_embeddings to match with inputs.
        """
        num_dim = len(inputs.shape)
        while num_dim - 1 > len(pos_embeddings.shape):
            pos_embeddings = pos_embeddings.unsqueeze(1)
        return pos_embeddings

    def forward(
        self,
        inputs: _torch.Tensor,
        *,
        positions: _Optional[_torch.Tensor] = None,
    ) -> _torch.Tensor:
        """
        Args:
            inputs (:obj:`torch.Tensor`): Input embeddings with shape
                ``(batch_size, sequence_len, dim)``.
            positions (:obj:`torch.Tensor`): Optional position integers with shape
                ``(batch_size, sequence_len)``.  Defaults to ``0, 1, ...`` for each
                sequence.
        """

        if self.sequence_start in (None, 0):
            start_inputs = None
        else:
            start_inputs = inputs[:, : self.sequence_start]
            inputs = inputs[:, self.sequence_start :]
            if positions is not None:
                positions = positions[:, self.sequence_start :]

        if positions is None:
            positions = _torch.arange(inputs.size(1), device=inputs.device)

        pos_embeddings = self.embedding(positions)
        pos_embeddings = self._maybe_broadcast_embedding(
            inputs=inputs, pos_embeddings=pos_embeddings
        )
        pos_embeddings = pos_embeddings.type_as(inputs)
        outputs = inputs + pos_embeddings

        if start_inputs is not None:
            return _torch.cat([start_inputs, outputs], dim=1)
        return outputs

    def extra_repr(self) -> str:
        if self.sequence_start in (None, 0):
            return ""
        return f"sequence_start={self.sequence_start}"


class SpatialPositionalEmbeddingCropMode(str, _enum.Enum):
    """
    Options for the crop behavior of :class`.SpatialPositionalEmbedding` layers.
    """

    #: Keep the top-left pieces of the weight parameter.
    TOP_LEFT = "TOP_LEFT"


class SpatialPositionalEmbedding(_nn.Module, _LayerMixin):
    """
    A layer for applying absolute positional embeddings to spatial inputs, such as
    NCHW image tensors.

    Args:
        spatial_shape (:obj:`tuple` of :obj:`int`): The spatial dimensions of the
            inputs.  For images, this is ``(height, width)``.
        dim (:obj:`int`): The embedding dimension (i.e., number of channels).
        interpolation (:obj:`.LayerBuilder` or :obj:`str`, optional): An
            :obj:`.Interpolation` layer for rescaling the spatial embeddings to match
            the shape of the layer's inputs.  If a :obj:`str`, this must be an
            interpolation mode, such as ``"bicubic"``.  Defaults to ``None``.
        interpolation_shape (:obj:`tuple` of :obj:`int`): The target size for
            interpolating positional embeddings.  Defaults to ``auto``.
            - If `"auto"`, the positional embedding weight is resized to match the input shape.
            - If a tuple is provided, the weight is resized to the given shape.
        input_features_dim (:obj:`int`): The index of the features (channels)
            dimension of the inputs.  This is ``1`` by default (for channels-first
            inputs).  Set this to ``-1`` for channels-last inputs.
        weight_crop_mode: (:obj:`str`, optional): Set to ``"top_left"`` to crop
            the (possibly interpolated) weight parameter when its size exceeds the size
            of the inputs.  Defaults to ``None``.
        device: The device for parameters.
        dtype: The dtype for parameters.
    """

    def __init__(
        self,
        spatial_shape: _Tuple[int, ...],
        dim: int,
        *,
        interpolation: _Union[_OptionalModuleOrBuilder, str] = None,
        interpolation_shape: _Union[str, _Tuple[int, ...]] = "auto",
        input_features_dim: int = 1,
        weight_crop_mode: _Optional[str] = None,
        device=None,
        dtype=None,
    ):
        super().__init__()

        self.spatial_shape = spatial_shape
        self.dim = dim

        if isinstance(interpolation, str):
            interpolation = _basic.Interpolation(mode=interpolation)
        self.interpolation = _helpers.maybe_build_module(interpolation)

        if interpolation_shape != "auto":
            interpolation_shape = tuple(interpolation_shape)
        self.interpolation_shape = interpolation_shape
        if self.interpolation is not None and self.interpolation_shape is None:
            raise ValueError(
                "interpolation_size should be specified when interpolation is needed"
            )

        self.input_features_dim = input_features_dim

        if weight_crop_mode is not None:
            weight_crop_mode = _helpers.get_enum_member_from_name(
                SpatialPositionalEmbeddingCropMode, weight_crop_mode
            )
        self.weight_crop_mode = weight_crop_mode

        weight_shape = (dim, *spatial_shape)
        self.weight = _torch.nn.Parameter(
            _torch.empty(weight_shape, device=device, dtype=dtype)
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        _init.shape_normalized_normal_(self.weight, dim=0)

    def extra_repr(self) -> str:
        return (
            f"shape={self.spatial_shape}, "
            f"dim={self.dim}, "
            f"interpolation_shape={self.interpolation_shape}, "
            f"weight_crop_mode={self.weight_crop_mode}"
        )

    def forward(self, inputs: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            inputs (:obj:`torch.Tensor`): Inputs with shape
                ``(batch_size, dim, *spatial_shape)``.

        Returns:
            The sum of the inputs and (possibly interpolated) positional embeddings.
        """

        weight = self.weight[None, ...]

        input_features_dim = self.input_features_dim % inputs.ndim
        inputs_spatial_shape = tuple(
            x
            for idx, x in enumerate(inputs.shape)
            if idx not in (0, input_features_dim)
        )

        if self.interpolation is not None:
            if (
                self.interpolation_shape == "auto"
                and weight.shape[2:] != inputs_spatial_shape
            ):
                weight = self.interpolation(weight, size=inputs_spatial_shape)
            elif weight.shape[2:] != self.interpolation_shape:
                weight = self.interpolation(weight, size=self.interpolation_shape)

        if self.weight_crop_mode is SpatialPositionalEmbeddingCropMode.TOP_LEFT:
            weight = tamm.layers.functional.crop(weight, shape=inputs_spatial_shape)

        if input_features_dim != 1:
            weight = weight.movedim(1, self.input_features_dim)

        return inputs + weight



import random as _random
from enum import Enum as _Enum
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn

from tamm.layers.transformer.kv_cache import BaseKVCache as _BaseKVCache
from tamm.layers.transformer.kv_cache import SpeculativeKVCache as _SpeculativeKVCache
from tamm.layers.transformer.kv_cache import V0KVCache as _V0KVCache
from tamm.layers.transformer.kv_cache import VanillaKVCache as _VanillaKVCache


def prepare_input_ids(input_ids: _torch.Tensor) -> _torch.Tensor:
    """Prepare and validate input tensor.

    Raises:
        ValueError: If input tensor has invalid dimensions.
    """
    input_ids = input_ids.clone().detach()
    if input_ids.ndim == 1:
        input_ids = input_ids.unsqueeze(0)

    if input_ids.ndim != 2:
        raise ValueError(
            f"Input IDs must be a 2D tensor (batch_size, seq_len). "
            f"Got {input_ids.ndim}D tensor."
        )

    return input_ids


def model_forward_step(
    model: _nn.Module,
    input_ids: _torch.Tensor,
    next_tokens: _Optional[_torch.Tensor],
    pad_id: _torch.Tensor,
    kv_cache: _Optional[_BaseKVCache],
    prompt_len: int,
    token_idx: _Optional[int] = 0,
) -> _torch.Tensor:
    segment_ids = input_ids.ne(pad_id)

    if kv_cache is None:
        return model(input_ids, segment_ids=segment_ids)
    if next_tokens is None:
        if isinstance(kv_cache, _V0KVCache):
            return model(
                input_ids,
                kv_cache=kv_cache.up_to_index(prompt_len),
            )
        return model(
            input_ids,
            segment_ids=segment_ids,
            kv_cache=kv_cache.up_to_index(prompt_len),
        )
    if token_idx is None:
        token_idx = 0
    return model(
        next_tokens,
        kv_cache=kv_cache.up_to_index(prompt_len + token_idx),
    )


def check_eos_endings(
    generated_tokens: _torch.Tensor, eos_token_ids: _List[_torch.Tensor]
) -> _torch.Tensor:
    # todo Do we have any EOS conditions?
    is_terminated = _torch.zeros(
        size=(generated_tokens.shape[0],),
        dtype=_torch.bool,
        device=generated_tokens.device,
    )
    for terminator in eos_token_ids:
        terminator_length = terminator.nelement()
        if generated_tokens.shape[1] >= terminator_length:
            endings = generated_tokens[..., -terminator_length:]
            is_terminated |= _torch.all(endings.eq(terminator), dim=1)
    return is_terminated


def find_eos_index(
    generated_tokens: _torch.Tensor, eos_token_ids: _List[_torch.Tensor]
) -> _torch.Tensor:
    termination_indices = _torch.full(
        size=(generated_tokens.shape[0],),
        dtype=_torch.int64,
        device=generated_tokens.device,
        fill_value=_torch.iinfo(_torch.int32).max,
    )

    for terminator in eos_token_ids:
        terminator_length = terminator.nelement()
        if generated_tokens.shape[1] >= terminator_length:
            windows = generated_tokens.unfold(
                1, terminator_length, 1
            )  # bs x seq_len - eos_len + 1 x eos_len
            matches = (windows == terminator).all(dim=2)  # bs x seq_len - eos_len + 1
            matches_any = matches.any(dim=1)
            first_occurence = matches.to(_torch.int32).argmax(dim=1)
            termination_indices[matches_any] = (
                first_occurence[matches_any] + terminator_length
            )
    return termination_indices


def process_logit(
    logits: _torch.Tensor,
    temperature: _Optional[float] = 1.0,
    top_k: _Optional[int] = None,
    top_p: _Optional[float] = None,
) -> _Tuple[_torch.Tensor, _torch.Tensor]:
    """Process raw logits from a decoder model. Applies temperature scaling, top-k and/or top-p (nucleus) sampling.

    Returns:
        :obj:`_Tuple[_torch.Tensor, _torch.Tensor]`: A tuple containing:
            - processed logits tensor
            - probability distribution vector derived from processed logits

    Raises:
        ValueError: If temperature <= 0, top_k <= 0, or top_p outside (0, 1].
    """
    # reshape to 2d matrix
    logits_shape = logits.shape
    vocab_size = logits_shape[-1]

    logits = logits.contiguous().view(-1, vocab_size)

    # Apply temperature scaling to next-token probability distribution
    if temperature is None:
        temperature = 1.0
    if temperature <= 0:
        raise ValueError(f"Invalid temperature value {temperature}, must be positive")
    logits = logits / temperature
    vocab_size = logits.size(-1)

    # Optionally select top-scoring tokens and ignore the rest
    # using top-K or top-P filtering
    if top_k is not None:
        if top_k <= 0:
            raise ValueError(f"Invalid top_k value {top_k}, must be positive")

        top_k = min(top_k, vocab_size)
        vals, indices_to_keep = _torch.topk(logits, top_k)

        # are we doing an unnecessary large allocation here?
        logits[:] = float("-inf")
        logits = logits.scatter(-1, indices_to_keep, vals)

    # skip top_p filtering if top_p = 1 because it does unnecessary
    # sorting
    probs = _torch.nn.functional.softmax(logits, dim=-1)
    if top_p is not None:
        if top_p <= 0 or top_p > 1.0:
            raise ValueError(f"Invalid top_p value {top_p}, must be between 0 and 1")

        min_tokens_to_keep = min(1, vocab_size)

        # Sort logits and compute (cumulative) probabilities
        sorted_probs, sorted_idxs = _torch.sort(probs, descending=False)
        cumulative_probs = sorted_probs.cumsum(dim=-1)

        # Remove tokens with cumulative probability exceeding threshold
        sorted_idxs_to_remove = cumulative_probs <= (1 - top_p)

        # Keep a minimum number of tokens
        sorted_idxs_to_remove[..., -min_tokens_to_keep:] = 0

        # Determine indices of unsorted tokens to keep
        idxs_to_remove = sorted_idxs_to_remove.scatter(
            -1, sorted_idxs, sorted_idxs_to_remove
        )

        # Over-write logits for tokens to filter out
        logits = logits.masked_fill(idxs_to_remove, -float("Inf"))

        # we zero-out the non-nucleus probabilities and re-normalize
        probs = probs.masked_fill(idxs_to_remove, 0.0)
        probs /= probs.sum(dim=-1, keepdim=True)

    # reshape back to 3d
    probs = probs.reshape(logits_shape)
    logits = logits.reshape(logits_shape)
    return logits, probs


class KVCacheType(_Enum):
    V0 = 1
    VANILLA = 2
    SPECULATIVE = 3


def setup_kv_cache(
    model: _nn.Module,
    batch_size: int,
    cache_len: int,
    input_ids: _torch.Tensor,
    pad_id: _Union[int, _torch.Tensor],
    cache_type: _Optional[KVCacheType] = None,
    cache_dtype: _Optional[_torch.dtype] = None,
) -> _Optional[_BaseKVCache]:
    """Setup KV cache for the model.

    Args:
        model (:obj:`nn.Module`): The model instance.
        batch_size (:obj:`int`): Batch size.
        cache_len (:obj:`int`): Cache length.
        input_ids (:obj:`torch.Tensor`): Input tensor.
        pad_id (:obj:`Union[int, torch.Tensor]`): Padding token ID.
        cache_type (:obj:`Optional[KVCacheType]`): Type of KV cache.
        cache_dtype (:obj:`Optional[torch.dtype]`): Data type for cache.

    Returns:
        :obj:`Optional[BaseKVCache]`: KV cache if enabled, None otherwise.
    """
    if cache_type is None:
        return None

    num_cache_layers = model.config.num_layers * getattr(model.config, "num_tracks", 1)

    try:
        kv_dim = model.kv_dim
    except AttributeError:
        kv_dim = (
            model.config.hidden_dim
            * model.config.num_kv_heads
            // model.config.num_heads
        )

    if isinstance(pad_id, _torch.Tensor):
        assert pad_id.dim() == 1
        pad_id = int(pad_id.flatten()[0].item())

    if cache_type == KVCacheType.V0:
        return _V0KVCache(
            layers=num_cache_layers,
            dim=kv_dim,
            batch=batch_size,
            init_seq=cache_len,
            offsets=_torch.argmax((input_ids != pad_id).int(), dim=-1),
            device=input_ids.device,
            dtype=cache_dtype,
        )
    if cache_type == KVCacheType.VANILLA:
        return _VanillaKVCache(
            num_layers=num_cache_layers,
            batch_size=batch_size,
            length=cache_len,
            hidden_dim=kv_dim,
            device=input_ids.device,
            dtype=cache_dtype,
        )
    if cache_type == KVCacheType.SPECULATIVE:
        return _SpeculativeKVCache(
            num_layers=num_cache_layers,
            batch_size=batch_size,
            length=cache_len,
            hidden_dim=kv_dim,
            device=input_ids.device,
            dtype=cache_dtype,
        )
    raise ValueError(f"Unrecognized cache type {cache_type}.")


def create_random_number_generator(
    device: _Union[str, _torch.device], seed: _Optional[int] = None
):
    """
    Create :py:class:`torch.Generator` on a specified device with a specified seed.
    """
    if seed is None:
        seed = _random.randint(0, 1_000_000)
    return _torch.Generator(device=device).manual_seed(seed)


def post_process_token_ids(
    token_ids: _torch.Tensor,
    max_new_tokens: _Union[int, _List[int]],
    pad_id: _Union[int, _torch.Tensor],
) -> _torch.Tensor:
    """
    Move all padding to the end of the sequence for all sequences in the batch and
    discard all extra tokens generated beyond max_new_tokens by setting them to pad
    """
    # cast mask to int because cuda cannot sort boolean tensor
    _, indices = _torch.sort((token_ids == pad_id).to(_torch.int32), dim=1, stable=True)
    token_ids = token_ids.gather(1, indices)
    if isinstance(max_new_tokens, list):
        max_max_new_tokens = max(max_new_tokens)
        max_new_tokens = (
            _torch.Tensor(max_new_tokens).unsqueeze(1).to(token_ids.device)
        )  # r x 1
        col_idx = _torch.arange(token_ids.shape[1], device=token_ids.device).unsqueeze(
            0
        )  # 1 x c
        mask = col_idx < max_new_tokens  # r x c
        token_ids = _torch.where(mask, token_ids, pad_id * _torch.ones_like(token_ids))
        token_ids = token_ids[:, :max_max_new_tokens]
    else:
        token_ids = token_ids[:, :max_new_tokens]
    return token_ids


def pad_after_termination(
    token_ids: _torch.Tensor,
    pad_id: _Union[int, _torch.Tensor],
    eos_token_ids: _List[_torch.Tensor],
):
    """
    Convert all token_ids greater than termination index to pad_id.
    """
    termination_indices = find_eos_index(token_ids, eos_token_ids)
    col_idx = _torch.arange(token_ids.shape[1], device=token_ids.device)
    mask = col_idx >= termination_indices[:, None]
    token_ids[mask] = pad_id
    return token_ids



from typing import ContextManager

from tamm.context_vars import model_build_device_context as _model_build_device_context
from tamm.layers.common.context_hooks.base import _BaseContextHook
from tamm.typing import OptionalDeviceOrString as _OptionalDeviceOrString


class DefaultDeviceContextHook(_BaseContextHook):
    def __init__(
        self,
        device: _OptionalDeviceOrString = None,
        disable_meta_device_init_trick: bool = False,
    ):
        """
        Args:
            device (:obj:`str`, optional):
                The new value for the ``torch`` default device. If ``None``,
                this will be a null context unless ``disable_meta_device_init_trick``
                is ``True``.
            disable_meta_device_init_trick (:obj:`bool`, optional):
                The flag indicating if the module should be built on actual devices.
                This is useful when we want to randomly initialize a specific module.
                If set to True, we will try to resolve the actual device that's in use.
                Defaults to ``False``.
        """
        self.device = device
        self.disable_meta_device_init_trick = disable_meta_device_init_trick

    def is_null(self) -> bool:
        return self.device is None and not self.disable_meta_device_init_trick

    def _get_context_manager(self) -> "ContextManager":
        return _model_build_device_context(
            device=self.device,
            disable_meta_init_trick=self.disable_meta_device_init_trick,
        )



"""
Preprocessor registry
=====================

This module defines baseclass for all ``PreprocessorRegistry``.

tamm plugin may extend Preprocessor registry by implementing a class that inherits :py:class:`.PreprocessorRegistry`
and publicize a custom Preprocessor registry by exposing a factory of PreprocessorRegistry via package metadata

.. code-block:: toml

    [project.entry-points.'tamm.plugins.extras']
    preprocessor-registry = 'module.submodule:get_preprocessor_registry'

"""
from tamm.preprocessors.registry.base import PreprocessorRegistry
from tamm.preprocessors.registry.uri_handler import URIHandlerPreprocessorRegistry

__all__ = ["PreprocessorRegistry", "URIHandlerPreprocessorRegistry"]



import contextlib as _contextlib
from contextvars import ContextVar

import torch

from tamm.context_vars._resolvers import _get_default_device
from tamm.typing import OptionalDeviceOrString

BUILD_DEVICE_NO_META_TRICK: ContextVar[OptionalDeviceOrString] = ContextVar(
    "build_device_no_meta_trick", default=None
)


def get_model_build_device(
    disable_meta_init_trick: bool = False,
) -> OptionalDeviceOrString:
    """
    A helper function for querying the current build device when using
    :func:`model_build_device_context`.

    Args:
        disable_meta_init_trick (:obj:`bool`): A flag for ignoring parent/ancestor
            usages of :func:`model_build_device_context` with
            ``enable_meta_init_trick=True``.  If ``False``, the function behaves
            as if these meta init contexts were never entered.
    """

    if disable_meta_init_trick:
        result = BUILD_DEVICE_NO_META_TRICK.get()
        if result is not None:
            return result

    return _get_default_device()


@_contextlib.contextmanager
def model_build_device_context(
    *,
    device: OptionalDeviceOrString = None,
    enable_meta_init_trick: bool = False,
    disable_meta_init_trick: bool = False,
):
    """
    A context manager similar to ``with torch.device(device)`` but with extra
    functionality for managing the "meta init trick".  (The meta init trick is
    when we build a pretrained model first on the torch meta device to avoid
    the cost of random initialization.  We then replace the meta parameters with
    real tensors that we load from a checkpoint.)

    Args:
        device (:obj:`str`, optional): The new value for the ``torch`` default
            device.  If ``None``, then we use the value of
            ``get_model_build_device(disable_meta_init_trick=disable_meta_init_trick)``,
            essentially keeping the value unchanged except for when the meta
            init trick is applied and we want to disable it. If
            ``enable_meta_init_trick`` is ``True``, then we force the device to "meta".
        enable_meta_init_trick (:obj:`bool`): Whether the purpose of the new
            context is specifically to apply the meta init trick.
        disable_meta_init_trick (:obj:`bool`): Whether to undo ancestor/parent
            applications of the context with ``enable_meta_init_trick=True``.  In
            this case, the default ``device`` becomes the value of
            ``get_model_build_device(disable_meta_init_trick=True)``.
    """

    if enable_meta_init_trick and disable_meta_init_trick:
        raise ValueError(
            "enable_meta_init_trick and disable_meta_init_trick cannot both be True"
        )

    if enable_meta_init_trick:
        device = "meta"
    if device is None:
        device = get_model_build_device(disable_meta_init_trick=disable_meta_init_trick)

    new_build_device_no_meta_trick = (
        get_model_build_device(disable_meta_init_trick=True)
        if enable_meta_init_trick
        else device
    )

    build_device_no_meta_trick_token = BUILD_DEVICE_NO_META_TRICK.set(
        new_build_device_no_meta_trick
    )

    try:
        with torch.device(device):
            yield
    finally:
        BUILD_DEVICE_NO_META_TRICK.reset(build_device_no_meta_trick_token)



from tamm.preprocessors.instruct_lm import afm, common
from tamm.preprocessors.instruct_lm.common import InstructLMPreprocessorOutput
from tamm.preprocessors.instruct_lm.jinja import JinjaChatTemplatePreprocessor

__all__ = [
    "afm",
    "common",
    "jinja",
    "JinjaChatTemplatePreprocessor",
    "InstructLMPreprocessorOutput",
]



import json as _json
from typing import Any as _Any
from typing import Callable as _Callable
from typing import List as _List

from tamm.utils.json._normalization import get_denormalizer as _get_denormalizer
from tamm.utils.json._normalization import get_normalizer as _get_normalizer
from tamm.utils.json._utils import CompositeDefault as _CompositeDefault
from tamm.utils.json._utils import CompositeObjectHooks as _CompositeObjectHooks
from tamm.utils.json._utils import DeviceJSONHook as _DeviceJSONHook
from tamm.utils.json._utils import DTypeJSONHook as _DTypeJSONHook
from tamm.utils.json._utils import global_default as _global_default
from tamm.utils.json._utils import global_object_hook as _global_object_hook

# tamm object lifecycle (from disk to to live object to disk):
#
#        |------------|  |-----------------------|
# JSON-> | normalizer |->| loads (de-serializer) |-----> tamm live object ----->
#        |------------|  |-----------------------|
#
#        |-------------------|  |--------------|
# -----> | dump (serializer) |->| denormalizer | -----> JSON
#        |-------------------|  |--------------|
#
# * normalizer  : To apply any JSON transformation (e.g., key replacement) prior to
#                 deserialization. Normalization can implement auto-update of
#                 JSON generated by older tamm versions.
#
# * denormalizer: To apply any JSON transformation after serialization. Denormalization
#                 can update JSON for compatibility with older tamm versions.
#


def loads(
    json_str: str,
    *,
    cls=None,
    object_hook=None,
    parse_float=None,
    parse_int=None,
    parse_constant=None,
    object_pairs_hook=None,
    **kw,
) -> _Any:
    """
    .. _Python standard json lib: https://docs.python.org/3/library/json.html
    .. _json.loads(): https://docs.python.org/3/library/json.html#json.loads

    A drop-in replacement of `json.loads()`_ that recognizes tamm objects

    Args:
        json_str: A json string representing tamm object or any supported object
        cls: See `Python standard json lib`_ for details
        object_hook: See `Python standard json lib`_ for details
        parse_float: See `Python standard json lib`_ for details
        parse_int: See `Python standard json lib`_ for details
        parse_constant: See `Python standard json lib`_ for details
        object_pairs_hook: See `Python standard json lib`_ for details

    Returns:
        A tamm object or any object that can be deserialized
    """
    json_str = _get_normalizer()(json_str)  # normalization for syntax compatibility
    _object_hooks: _List[_Callable[..., _Any]] = [
        _DeviceJSONHook().object_hook,
        _DTypeJSONHook().object_hook,
        _global_object_hook,
    ]
    if object_hook is not None:
        _object_hooks.append(object_hook)

    _composite_hook = _CompositeObjectHooks(*_object_hooks)
    return _json.loads(
        json_str,
        cls=cls,
        object_hook=_composite_hook,
        parse_float=parse_float,
        parse_int=parse_int,
        parse_constant=parse_constant,
        object_pairs_hook=object_pairs_hook,
        **kw,
    )


def load(fp, **kwargs):
    """
    A drop-in replacement for :func:`json.load` that recognizes |tamm| objects.

    Args:
        fp: A file-like object (with a :meth:`read` method) containing the JSON
            document.
        **kwargs: Keyword arguments for :func:`json.loads`.

    Returns:
        The deserialized object.
    """
    return loads(fp.read(), **kwargs)


def dumps(
    obj: _Any,
    *,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    indent=None,
    separators=None,
    default=None,
    sort_keys=False,
    **kw,
) -> str:
    """
    .. _json.dumps(): https://docs.python.org/3/library/json.html#json.dumps

    A drop-in replacement of `json.dumps()`_ that recognizes tamm objects

    Args:
        obj: tamm object or any object
        skipkeys: See `Python standard json lib`_ for details
        ensure_ascii: See `Python standard json lib`_ for details
        check_circular: See `Python standard json lib`_ for details
        allow_nan: See `Python standard json lib`_ for details
        cls: See `Python standard json lib`_ for details
        indent: See `Python standard json lib`_ for details
        separators: `Python standard json lib`_ for details
        default: `Python standard json lib`_ for details
        sort_keys: See `Python standard json lib`_ for details

    Returns:
        A JSON string literal representing a serialized object

    """
    _defaults = [_global_default, _DTypeJSONHook().default, _DeviceJSONHook().default]
    if default is not None:
        _defaults.append(default)

    _composite_hook = _CompositeDefault(*_defaults)
    kwargs = {
        "skipkeys": skipkeys,
        "ensure_ascii": ensure_ascii,
        "check_circular": check_circular,
        "allow_nan": allow_nan,
        "cls": cls,
        "indent": indent,
        "separators": separators,
        "sort_keys": sort_keys,
    }
    kwargs.update(kw)
    output_json = _json.dumps(
        obj,
        default=_composite_hook,
        **kwargs,
    )
    return _get_denormalizer()(
        output_json,
        **kwargs,
    )


def dump(obj, fp, **kwargs):
    """
    A drop-in replacement for :func:`json.dump` that recognizes |tamm| objects.

    Args:
        fp: A file-like object (with a :meth:`write` method).
        **kwargs: Keyword arguments for :func:`json.dump`.
    """
    fp.write(dumps(obj, **kwargs))



from pathlib import Path
from typing import Union

import tamm.utils.json as tamm_json
from tamm.model_repo.exceptions import UnrecognizedModelIdentifierError
from tamm.model_repo.model_repo import BaseModelRepo
from tamm.model_repo.utils import get_model_config_from_any_tamm_object


class FileModelRepo(BaseModelRepo):
    def _get_configs_impl(self):
        return {}

    def clear_cache(self):
        ...

    def is_alive(self) -> bool:
        return True

    def create_model_config(self, model_name: Union[str, Path], *args, **kwargs):
        if not isinstance(model_name, (str, Path)):
            raise UnrecognizedModelIdentifierError(
                f"model_name must be str, Path, not {type(model_name).__name__}"
            )
        if isinstance(model_name, str) and model_name.lower().startswith("file://"):
            core_model_name = model_name[len("file://") :]
            if not Path(core_model_name).is_file():
                raise FileNotFoundError(f"No such file: {model_name}")
            model_name = core_model_name

        if Path(model_name).is_file():
            config_path = Path(model_name)
        else:
            raise UnrecognizedModelIdentifierError(f"{self} cannot create {model_name}")

        with open(config_path, "r", encoding="utf-8") as fptr:
            maybe_model_config = tamm_json.load(fptr)

        model_config = get_model_config_from_any_tamm_object(maybe_model_config)
        model_config.update_configured_args(*args, **kwargs)
        return model_config



import dataclasses as _dataclasses
import enum as _enum
from typing import Dict as _Dict
from typing import Iterable, List, Union

from tamm._helpers import case_insensitive_lookup as _case_insensitive_lookup
from tamm.utils import partial as _partial
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin

# pylint: disable=duplicate-code


class TextProcessorConfig(
    _partial.DataclassedPartial,
    _JSONSerializableMixin,
    json_namespace="text_processors",
):
    # pylint:disable=useless-parent-delegation
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_text_processor(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured instance of the text processor.

        Args:
            override_args: Optional positional arguments to override args specified in
                the config.  These args replace the first ``len(override_args)``
                positional args.
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overriden by
                ``override_args``.

        Returns:
            The newly created tokenizer.
        """
        return self(*override_args, **override_kwargs)

    def _to_json_dict_impl(self):
        result = {}
        for field in _dataclasses.fields(self.configured_args):
            value = getattr(self.configured_args, field.name)
            if value == field.default:
                continue
            result[field.name] = value
        return result


class TextProcessor:
    """
    Text processor class for pre-processing methods to be applied on a text before
    calling Tokenizer.
    """

    def _process(self, text: str) -> str:
        """
        Apply pre-processing on a single piece of text.
        """
        raise NotImplementedError()

    def process(self, texts: Union[str, Iterable[str]]) -> Union[str, List[str]]:
        """
        Apply pre-processing on a single piece of text or an iterable of texts.
        """
        if isinstance(texts, str):
            return self._process(texts)

        if issubclass(type(texts), Iterable):
            return [self._process(text) for text in texts]

        raise TypeError("Input must be a string or a list of strings")

    def _inverse(self, text: str) -> str:
        """
        Revert the effects of pre-processing on a single piece of text.
        """
        raise NotImplementedError()

    def inverse(self, texts: Union[str, Iterable[str]]) -> Union[str, List[str]]:
        """
        Revert the effects of pre-processing on a single piece of text or an
        iterable of texts.
        """
        if isinstance(texts, str):
            return self._inverse(texts)

        if issubclass(type(texts), Iterable):
            return [self._inverse(text) for text in texts]

        raise TypeError("Input must be a string or an iterable of strings")

    def __init_subclass__(cls):
        text_processor_name = cls.__name__
        config_name = f"{text_processor_name}Config"
        config_cls = TextProcessorConfig.create_subclass(
            target_callable=cls, name=config_name
        )
        config_cls.__doc__ = (
            f"A :py:class:`.TextProcessorBase` subclass for configuring "
            f":py:class:`.{text_processor_name}` text processor. "
            f"Use the alias :attr:`.{text_processor_name}.Config` to access this class. "
            f"Please check :class:`.{text_processor_name}` for more details about the "
            "signature."
        )
        cls.Config = config_cls


class FindAndReplaceTextProcessor(TextProcessor):
    """
    Handles text preprocessing by replacing substrings with specified replacements.
    """

    def __init__(self, replacements: _Dict[str, str]):
        self.replacements = dict(replacements)

    def _process(self, text: str) -> str:
        """Applies text replacements on a given text."""
        for search, repl in self.replacements.items():
            text = text.replace(search, repl)
        return text

    def _inverse(self, text: str) -> str:
        """Revert the effects of pre-processing on a given text."""
        for original, repl in reversed(self.replacements.items()):
            text = text.replace(repl, original)
        return text


class TextProcessorMode(str, _enum.Enum):
    #: FindAndReplaceTextProcessor with replacements={"\n": "<n>"}
    AFM_TEXT_FIND_AND_REPLACE = "AFM_TEXT_FIND_AND_REPLACE"

    @classmethod
    def _missing_(cls, value):
        return _case_insensitive_lookup(cls, value)



import tamm.utils.json as _tamm_json
from tamm.model_repo.exceptions import UnrecognizedModelIdentifierError
from tamm.model_repo.model_repo import BaseModelRepo
from tamm.model_repo.utils import get_model_config_from_any_tamm_object
from tamm.utils.uri import _URIHandler


class URIHandlerModelRepo(BaseModelRepo):
    def _get_configs_impl(self):
        return {}

    def clear_cache(self):
        ...

    def is_alive(self) -> bool:
        return True

    def create_model_config(self, model_name: str, *args, **kwargs):
        """
        Use _URIHandler to load model configs from any supported URI

        Args:
            model_name: URI like s3://bucket/tamm-config.json
            *args: model config arguments
            **kwargs: model config keyword arguments

        Returns: ModelConfig subclass instance

        """
        if not isinstance(model_name, str):
            raise UnrecognizedModelIdentifierError(
                f"model_name must be a string, got {type(model_name)}"
            )
        try:
            with _URIHandler().open(model_name, mode="r") as f:
                maybe_model_config = _tamm_json.load(f)
        except FileNotFoundError as e:
            # cfassetlib raises FileNotFoundError
            raise FileNotFoundError(f"No such file: {model_name}") from e
        except Exception as e:
            raise UnrecognizedModelIdentifierError(
                f"{self} cannot create {model_name}"
            ) from e

        model_config = get_model_config_from_any_tamm_object(maybe_model_config)
        model_config.update_configured_args(*args, **kwargs)
        return model_config



from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm import _helpers
from tamm.layers import functional as _tamm_F
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.mixture_of_experts.dispatcher import common as _common
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder

# pylint: disable=duplicate-code


class CapacityConstrainedMixtureOfExpertsDispatcher(
    _common.MixtureOfExpertsDispatcher, _LayerMixin
):
    """
    A dispatcher that transforms inputs of shape
    ``(batch_size, seq_len, hidden_dim)`` into expert inputs with shape
    ``(num_experts, expert_capacity, hidden_dim)``.  A capacity constraint
    child layer determines the size of ``expert_capacity``. The dispatcher
    drops tokens when the number of tokens for an expert exceeds the capacity.

    This dispatcher is intended for experts implemented using batched matrix
    multiplies.

    The layer also supports multiple sets of experts (each set of size
    ``num_experts``) via the ``track_dim`` argument.  For this, the layer
    maps inputs with shape ``(num_tracks, batch_size, seq_len, hidden_dim)``
    to an experts input with shape
    ``(num_tracks, num_experts, expert_capacity, hidden_dim)``.

    Args:
        num_experts (:obj:`int`): The number of experts.  If ``track_dim``
            is not ``None``, this is the number of experts per track.
        capacity_constraint: A builder for a
            :obj:`MixtureOfExpertsCapacityConstraint` child layer.
        track_dim (:obj:`int`, optional): If using multiple sets (tracks) of
            experts, this is the index of the track dimension in the
            :meth:`forward` dispatching inputs.  Defaults to ``None``.
    """

    def __init__(
        self,
        num_experts: int,
        capacity_constraint: _ModuleOrBuilder,
        track_dim: _Optional[int] = None,
    ):
        super().__init__()
        self.num_experts = num_experts
        _helpers.append_children(self, capacity_constraint=capacity_constraint)
        self.track_dim = track_dim

    def extra_repr(self):
        components = [f"num_experts={self.num_experts}"]
        if self.track_dim is not None:
            components.append(f"track_dim={self.track_dim}")
        return ", ".join(components)

    @property
    def _num_track_dims(self):
        return 0 if self.track_dim is None else 1

    def _dispatch(
        self, hidden_states: _torch.Tensor, *, expert_assignments: _torch.Tensor
    ) -> _common.MixtureOfExpertsDispatchOutput:
        # shorthand for shape comments:
        # - T: num tracks
        # - B: batch size
        # - S: sequence length
        # - D: hidden dim
        # - K: experts per token
        # - E: num experts
        # - C: capacity per expert

        x = hidden_states

        if self.track_dim not in (0, None):
            x = x.movedim(self.track_dim, 0)  # (T, B, S, D)
            expert_assignments = expert_assignments.movedim(
                self.track_dim, 0
            )  # (T, B, S, K)

        one_hots = _torch.nn.functional.one_hot(  # pylint: disable=not-callable
            expert_assignments, self.num_experts
        )  # (T, B, S, K, E)

        constraint_out = self.capacity_constraint(
            one_hots, num_track_dims=self._num_track_dims
        )
        capacity_per_expert: int = constraint_out.capacity_per_expert
        capacity_per_expert = min(
            capacity_per_expert,
            expert_assignments.shape[self._num_track_dims :].numel(),
        )  # cap to total number of tokens

        capacity_indices = self._compute_capacity_indices(
            assignment_indicators=constraint_out.assignment_indicators,
            capacity_per_expert=capacity_per_expert,
        )  # (T, B, S, K, E)

        dispatch_indices = self._compute_dispatch_indices_from_capacity_indices(
            capacity_indices=capacity_indices,
            capacity_per_expert=capacity_per_expert,
        )  # (T, E*C)

        collapsed_indices = capacity_indices.max(dim=-1)[0]  # (T, B, S, K)
        dropped_assignments = collapsed_indices == -1
        combine_indices = (
            expert_assignments * capacity_per_expert + collapsed_indices
        )  # (T, B, S, K)
        combine_indices = combine_indices.masked_fill(dropped_assignments, 0)

        num_experts_per_token = expert_assignments.size(-1)
        x = x.expand(num_experts_per_token, *x.shape)  # (K, T, B, S, D)
        x = x.movedim(0, -2)  # (T, B, S, K, D)

        x = x.flatten(start_dim=self._num_track_dims, end_dim=-2)  # (T, B*S*K, D)
        experts_arg = _torch.take_along_dim(
            x,  # (T, B*S*K, D)
            dispatch_indices.unsqueeze(-1),  # (T, E*C, 1)
            dim=-2,
        )  # (T, E*C, D)
        experts_arg = experts_arg.unflatten(-2, (self.num_experts, -1))  # (T, E, C, D)

        return _common.MixtureOfExpertsDispatchOutput(
            experts_arg=experts_arg,
            experts_kwargs={},
            combine_input={
                "combine_indices": combine_indices,
                "dropped_assignments": dropped_assignments,
            },
        )

    def _compute_capacity_indices(self, *, assignment_indicators, capacity_per_expert):
        """
        Args:
            assignment_indicators (:obj:`torch.Tensor`): A tensor with shape
                ``(*track_shape, *batch_shape, num_experts_per_token, num_experts)``.
                A value of ``1`` indicates a token assigned to an expert.
            capacity_per_expert (:obj:`int`): The capacity per expert.

        Returns:
            An integer :obj:`torch.Tensor` with the same shape as the inputs.
            The tensor contains each token's capacity index for its expert.
            Values of ``-1`` indicate dropped tokens or tokens not assigned to an
            expert.
        """
        bsk_shape = assignment_indicators.shape[self._num_track_dims : -1]
        assignment_indicators = assignment_indicators.flatten(
            start_dim=self._num_track_dims, end_dim=-2
        )  # (T, B*S*K, E)
        indices = _tamm_F.cumsum(assignment_indicators, dim=-2)  # (T, B*S*K, E)
        indices *= assignment_indicators
        indices *= indices <= capacity_per_expert
        indices -= 1
        return indices.unflatten(-2, bsk_shape)  # (T, B, S, K, E)

    def _compute_dispatch_indices_from_capacity_indices(
        self, capacity_indices, capacity_per_expert
    ):
        """
        Args:
            capacity_indices: Integer tensor with shape (T, B, S, K, E).
                A value of -1 represents a non-token or dropped token.  A
                value >= 0 represents the capacity index for a token.
            capacity_per_expert: The max number of tokens to assign that
                each expert will receive.
        """
        capacity_indices = _torch.masked_fill(
            capacity_indices, capacity_indices < 0, capacity_per_expert
        )  # (T, B, S, K, E)
        capacity_indices = capacity_indices.flatten(
            start_dim=self._num_track_dims, end_dim=-2
        )  # (T, B*S*K, E)
        values, dispatch_indices = capacity_indices.topk(
            capacity_per_expert, dim=-2, largest=False
        )  # (T, C, E)
        dispatch_indices = dispatch_indices.masked_fill_(
            values >= capacity_per_expert, 0
        )  # (T, C, E)
        dispatch_indices = dispatch_indices.transpose(-1, -2)  # (T, E, C)
        return dispatch_indices.flatten(start_dim=-2)  # (T, E*C)

    def _combine(
        self,
        experts_output: _torch.Tensor,
        *,
        assignment_weights: _torch.Tensor,
        combine_input: _Dict[str, _torch.Tensor],
    ) -> _torch.Tensor:
        if self.track_dim not in (0, None):
            assignment_weights = assignment_weights.movedim(
                self.track_dim, 0
            )  # (T, B, S, K)

        combine_indices = combine_input["combine_indices"]
        dropped_assignments = combine_input["dropped_assignments"]

        experts_output = experts_output.flatten(
            start_dim=self._num_track_dims, end_dim=-2
        )  # (T, E*C, D)
        bsk_shape = combine_indices.shape[self._num_track_dims :]
        combine_indices = combine_indices.flatten(
            start_dim=self._num_track_dims
        )  # (T, B*S*K)
        x = _torch.take_along_dim(
            experts_output,  # (T, E*C, D)
            combine_indices.unsqueeze(-1),  # (T, B*S*K, 1)
            dim=-2,
        )  # (T, B*S*K, D)
        x = x.unflatten(-2, bsk_shape)  # (T, B, S, K, D)

        assignment_weights = _torch.masked_fill(
            assignment_weights, dropped_assignments, 0
        )
        assignment_weights = assignment_weights.to(x.dtype)

        x = _torch.linalg.vecdot(  # pylint: disable=not-callable
            x, assignment_weights.unsqueeze(-1), dim=-2
        )  # (T, B, S, D)

        if self.track_dim not in (0, None):
            x = x.movedim(0, self.track_dim)

        return x



import abc
import inspect
from typing import Any

from tamm.utils.json import JSONSerializableABCMixin


class Preprocessor(JSONSerializableABCMixin, json_namespace="preprocessor"):
    """
    Abstract base class for all preprocessors

    The signature of init function will become the serialized json 'config'.
    All __init__ arguments must be saved as instance attributes for JSON serialization
    to work.

    For example,

    .. code-block:: python

        class MyPreprocessor(_Preprocessor):

            def __init__(
                self,
                attribute1 = "value1"
                attribute2 = "value2"
            ):
                self.attribute1 = attribute1 # required for json serialization
                self.attribute2 = attribute2 # required for json serialization

    will serialize into

    .. code-block:: json

        {
            "__tamm_type__": "preprocessor:MyPreprocessor",
            "attribute1": "value1",
            "attribute2": "value1",
        }

    """

    @abc.abstractmethod
    def __call__(self, *example: Any) -> Any:
        """
        Process one example

        Args:
            example: Raw data of any Python class which can be preprocessed by concrete
                     preprocessor

        Returns: Any preprocessed object
        """

    def __eq__(self, other):
        if other is None:
            return False
        return self.__dict__ == other.__dict__

    def _to_json_dict_impl(self) -> dict:
        """
        Default serializer implementation of all preprocessor.
        Serializes all ``__init__`` arguments, which are essentially preprocessor
        config, to dict.

        Assumes all __init__ arguments be kept track of, as instance attributes
        """
        init_signature = inspect.signature(self.__init__)  # type: ignore[misc]
        init_arguments = list(init_signature.parameters.items())
        init_arguments = [
            (k, v)
            for k, v in init_arguments
            if v.kind not in {v.VAR_KEYWORD, v.VAR_POSITIONAL}
        ]
        if not init_arguments:
            return {}
        config_attrs = {}
        for p, _ in init_arguments:
            try:
                config_attrs[p] = getattr(self, p)
            except AttributeError as e:
                raise TypeError(
                    f"Cannot JSON serialize '{self.__class__.__name__}' "
                    f"instance because __init__  argument '{p}' isn't kept track of "
                    f"as instance attribute"
                ) from e
        return self._omit_defaults(config_attrs)

    def _omit_defaults(self, attr_dict: dict) -> dict:
        attr_dict = attr_dict.copy()
        init_signature = inspect.signature(self.__init__)  # type: ignore[misc]
        attrs_with_default_value = []
        for k, v in attr_dict.items():
            try:
                if v == init_signature.parameters[k].default:
                    attrs_with_default_value.append(k)
            except KeyError:
                pass
        for attr_to_remove in attrs_with_default_value:
            attr_dict.pop(attr_to_remove)
        return attr_dict

    @classmethod
    def _from_json_dict_impl(cls, **raw_dict) -> "Preprocessor":
        """
        De-serializes all attributes from dict by passing them through ``__init__``

        Args:
            **raw_dict: Dictionary representation (i.e., config)
                        of this preprocessor

        Returns: ``Preprocessor``

        """
        new_instance = cls(**raw_dict)
        return new_instance



"""
URI Handlers
------------

|tamm| includes built-in URI handlers to facilitate checkpoint downloading with disk caching.
To support custom storage backends, plugin may extend URI handlers using the following process:

Register a custom URI Handler
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Create a submodule named ``<plugin>.utils.uri`` and add your URI handler to this submodule.
For example, to implement ``Boto3URIHandler``, create a ``boto3.py``  under ``<my_tamm_plugin>/utils/uri``
with the following content:

.. code-block:: python

    from tamm.utils.uri._protocol import BaseURLHandler
    from pathlib import Path

    class Boto3URIHandler(BaseURLHandler):
        name = "boto3"

        def _map_to_local(self, uri: str, local_path: Path) -> Path:
            # Implement downloading w/ boto3. Return the downloaded (or cached) local path
            return local_path

.. admonition:: Tip

    submodule naming must be exactly ``<plugin>.utils.uri`` for the plugin system to recognize URI handlers

Register your URI handler by exporting ``available_handlers`` from ``<plugin>/utils/uri/__init__.py``.
``available_handlers`` should be a list of ``BaseURLHandler``, specifically, ``List[Type["BaseURLHandler"]]``

.. code-block:: python

    from my_tamm_plugin.utils.uri.boto3 import Boto3URIHandler

    available_handlers = [Boto3URIHandler]
    __all__ = [
        "Boto3URIHandler",
        "available_handlers",
    ]

.. admonition:: Tip

    Import statement ``from my_tamm_plugin.utils.uri import available_handlers`` **must** work for
    |tamm| plugin system to discover your custom URI handler.

URI Handler Base Classes
^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: tamm.utils.uri.BaseURLHandler
.. autoclass:: tamm.utils.uri.AtomicDownloadURIHandler

"""

from tamm.utils.uri._protocol import AtomicDownloadURIHandler, BaseURLHandler
from tamm.utils.uri.uri_handler import _is_json_file, _is_uri, _URIHandler

__all__ = [
    "_URIHandler",
    "_is_uri",
    "_is_json_file",
    "BaseURLHandler",
    "AtomicDownloadURIHandler",
]



"""
Implements a |tamm| tokenizer using SentencePiece.
"""

import logging
import os
from functools import cached_property
from typing import Any, Iterable, List, Optional, Union, cast

from packaging.version import parse as parse_version

from tamm.tokenizers.common import Tokenizer as _TokenizerBase
from tamm.tokenizers.text_processors import TextProcessor as _TextProcessorBase

logger = logging.getLogger(__name__)


def _add_user_defined_tokens_and_save_to_file(
    vocab_path, output_path, user_defined_tokens
):
    os.environ[
        "PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"
    ] = "python"  # Force protobuf to use Python backend, will not affect parent process

    # pylint: disable=import-outside-toplevel,import-error
    from sentencepiece import SentencePieceProcessor

    tokenizer = SentencePieceProcessor()
    tokenizer.Load(vocab_path)
    logger.debug(
        f"Loaded vocab file from {vocab_path}, vocab size = {tokenizer.vocab_size()}"
    )
    _add_user_defined_tokens(tokenizer, user_defined_tokens)
    with open(output_path, "wb") as file:
        file.write(tokenizer.serialized_model_proto())
        logger.debug(
            f"New vocab size = {tokenizer.vocab_size()}, written to {output_path}"
        )


def _add_user_defined_tokens_in_new_process(
    tokenizer: Any, vocab_path: str, user_defined_tokens: Iterable[str]
):
    # pylint: disable=import-outside-toplevel
    import multiprocessing as mp
    import tempfile

    from sentencepiece import SentencePieceProcessor  # pylint: disable=import-error

    assert isinstance(tokenizer, SentencePieceProcessor)
    with tempfile.NamedTemporaryFile() as temp_file:
        # First, pad the vocab in a new process and save to `temp_file`
        logger.debug("Starting a new process to add user defined tokens")
        ctx = mp.get_context("spawn")
        p = ctx.Process(
            target=_add_user_defined_tokens_and_save_to_file,
            args=(vocab_path, temp_file.name, user_defined_tokens),
        )
        p.start()
        p.join()
        if p.exitcode != 0:
            # pylint: disable=raise-missing-from
            raise RuntimeError(
                "Failed to pad sentencepiece model,"
                f" subprocess failed with exit code {p.exitcode}"
            )
        logger.debug("The new process exited successfully")

        # Load updated vocab from `temp_file`
        tokenizer.Load(temp_file.name)
        logger.debug("Loaded updated vocab file")


def _add_user_defined_tokens(tokenizer: Any, user_defined_tokens: Iterable[str]):
    # pylint: disable=import-outside-toplevel,import-error
    from sentencepiece import SentencePieceProcessor
    from sentencepiece import sentencepiece_model_pb2 as sp_pb2

    assert isinstance(tokenizer, SentencePieceProcessor)
    pb = sp_pb2.ModelProto()  # pylint: disable=invalid-name, E1101
    pb.ParseFromString(tokenizer.serialized_model_proto())
    for token in user_defined_tokens:
        # Add an user defined token
        new_token = sp_pb2.ModelProto().SentencePiece()  # pylint: disable=E1101
        new_token.piece = token
        new_token.score = 0.0
        new_token.type = 4  # USER_DEFINED
        pb.pieces.append(new_token)
    tokenizer.LoadFromSerializedProto(pb.SerializeToString())


class SentencePieceTokenizer(_TokenizerBase):
    """
    Base class for SentencePiece Tokenizers.

    Args:
        vocab_path (:obj:`str`): Path to sentencepiece vocab file.
        user_defined_tokens (:obj:`list`, optional): List of user defined tokens to
            add to vocab. Defaults to `None`.
        pad_vocab_size_to (:obj:`int`, optional): Pad the current vocab size to this
            size. If this value is not `None`, ..., "<extra_id_2>", "<extra_id_1>",
            "<extra_id_0>" will be added to vocab, otherwise no padding will be added.
            Defaults to `None`.
        text_processor(:obj:`tamm.tokenizers.base.TextProcessorBase`, optional): Text
            process to use before encoding and after decoding. Defaults to `None`.
    """

    def __init__(
        self,
        vocab_path: str,
        user_defined_tokens: Optional[Iterable[str]] = None,
        pad_vocab_size_to: Optional[int] = None,
        text_processor: Optional[_TextProcessorBase] = None,
    ):
        try:
            # lazily import sentencepiece to support using tamm without sentencepiece
            # pylint: disable=import-outside-toplevel
            import sentencepiece
        except ImportError:
            logger.error(
                "tamm tried to initialize a tokenizer that requires sentencepiece, but "
                "tamm could not import this.  Please install sentencepiece or check "
                "your installation of this package."
            )
            raise

        self._vocab_path = vocab_path
        self._tokenizer = sentencepiece.SentencePieceProcessor()
        self._tokenizer.Load(self._vocab_path)
        self.text_processor = text_processor

        additional_tokens = []
        if user_defined_tokens is not None:
            additional_tokens += user_defined_tokens

        if pad_vocab_size_to is not None:
            additional_tokens += [
                f"<extra_id_{idx}>"
                for idx in reversed(
                    range(pad_vocab_size_to - len(self) - len(additional_tokens))
                )
            ]

        if len(additional_tokens) > 0:
            self.add_user_defined_tokens(additional_tokens)

    def add_user_defined_tokens(self, user_defined_tokens: Iterable[str]):
        # pylint: disable=import-outside-toplevel,unused-import
        try:
            import sentencepiece  # noqa: F401
        except ImportError:
            logger.error(
                "tamm tried to initialize a tokenizer that requires sentencepiece, but "
                "tamm could not import this.  Please install sentencepiece or check "
                "your installation of this package."
            )
            raise

        from google import protobuf

        # Dispatch calls based on protobuf version
        if parse_version(protobuf.__version__) < parse_version("4.0"):
            _add_user_defined_tokens(self._tokenizer, user_defined_tokens)
        else:
            logger.debug(
                "Fall back to Python protobuf backend to add user defined tokens"
                f" because protobuf version {protobuf.__version__} is not supported"
            )
            _add_user_defined_tokens_in_new_process(
                self._tokenizer, self.vocab_path, user_defined_tokens
            )
        logger.debug(f"Added the following tokens {user_defined_tokens}")

    def __len__(self) -> int:
        """Vocabulary size of the SentencePiece tokenizer."""
        return cast(int, self.tokenizer.vocab_size())

    @property
    def vocab_path(self):
        return self._vocab_path

    @property
    def tokenizer(self):
        """The SentencePiece tokenizer."""
        return self._tokenizer

    def encode(
        self,
        texts: Union[str, Iterable[str]],
        apply_processor: bool = True,
        **kwargs,  # pylint: disable=arguments-differ
    ) -> Union[List[int], List[List[int]]]:
        """Converts a string or a list of strings into a list of token IDs."""
        if apply_processor and self.text_processor is not None:
            texts = self.text_processor.process(texts)
        return self.tokenizer.Encode(texts, **kwargs)

    def decode(  # pylint: disable=arguments-differ
        self,
        token_ids: Union[int, Iterable[int], Iterable[Iterable[int]]],
        apply_processor: bool = True,
        **kwargs,
    ) -> Union[str, List[str]]:
        """Decodes token IDs to their corresponding strings."""
        decoded_texts = self.tokenizer.Decode(token_ids, **kwargs)
        if apply_processor and self.text_processor is not None:
            decoded_texts = self.text_processor.inverse(decoded_texts)
        return decoded_texts

    def encode_as_pieces(  # pylint: disable=arguments-differ
        self,
        texts: Union[str, Iterable[str]],
        apply_processor: bool = True,
    ) -> Union[List[str], List[List[str]]]:
        """Tokenizes texts into subword pieces."""
        if apply_processor and self.text_processor is not None:
            texts = self.text_processor.process(texts)
        return cast(List[str], self.tokenizer.EncodeAsPieces(texts))

    @cached_property
    def pad_id(self) -> int:
        return cast(int, self.tokenizer.pad_id())

    @cached_property
    def eos_id(self) -> int:
        return cast(int, self.tokenizer.eos_id())

    @cached_property
    def unk_id(self) -> int:
        return cast(int, self.tokenizer.unk_id())



import contextlib
from typing import List as _List
from typing import Type as _Type
from typing import Union as _Union

from torch import nn as _nn

from tamm.ao.layers import FakeQuantize


def get_fake_quantize_layers(
    model: _nn.Module, parent_modules: _Union[None, _Type, _List[_Type]] = None
):
    """
    This method compiles all `FakeQuantize` layers which have a parent module in
    `model` in `parent_modules`, returning a dictionary of module names to
    `FakeQuantize` layers. If `parent_modules` is `None`, all `FakeQuantize` layers
    are returned.

    Args:
        model (:obj:`_nn.Module`): The model to get `FakeQuantize` layers from.
        parent_modules: (:obj:`Type` | `List[Type]`): Parent module class(es) to look
            for `FakeQuantize` layers in.  For example, if set to `TransformerAttention`,
            only `FakeQuantize` layers in `TransformerAttention` modules are returned.
            Defaults to `None` which does not perform any filtering.
    """
    layers = {}
    if isinstance(parent_modules, list):
        parent_modules = tuple(parent_modules)
    for name, module in model.named_modules():
        if parent_modules is None and isinstance(module, FakeQuantize):
            layers[name] = module
        elif parent_modules is not None and isinstance(module, parent_modules):
            layers_from_child = get_fake_quantize_layers(module, parent_modules=None)
            for child_name, child_module in layers_from_child.items():
                new_name = name
                if len(child_name) > 0:
                    new_name = f"{new_name}.{child_name}"
                layers[new_name] = child_module
    return layers


def set_fake_quantize_enabled(
    model: _nn.Module,
    value: bool,
    parent_modules: _Union[None, _Type, _List[_Type]] = None,
):
    """
    This method sets all `FakeQuantize.is_fake_quant_enabled` to `value` if a parent
    module in `model` is in `parent_modules`. If `parent_modules` is `None`, this is
    applied to all `FakeQuantize` layers.

    Args:
        model (:obj:`_nn.Module`): Model to set `FakeQuantize.is_fake_quant_enabled`.
        parent_modules: (:obj:`Type` | `List[Type]`): Parent module class(es) to set
            `FakeQuantize.is_fake_quant_enabled` layers in. For example, if set to
            `TransformerAttention`, only `FakeQuantize` layers in `TransformerAttention`
            modules are accessed. Defaults to `None` which does not filter.
    """
    layers = get_fake_quantize_layers(model, parent_modules=parent_modules)
    for fq in layers.values():
        fq.is_fake_quant_enabled = value


@contextlib.contextmanager
def fake_quantize_disabled_context(
    model: _nn.Module, parent_modules: _Union[None, _Type, _List[_Type]] = None
):
    """
    This context disables the quantize and dequantization steps in all `FakeQuantize`
    layers in `model` i.e. sets `FakeQuantize.is_fake_quant_enabled` to `False`. This
    does not affect `FakeQuantize.activation_post_process`, so quantization parameters
    may still update if enabled.

    Args:
        model (:obj:`_nn.Module`): The model to disable `FakeQuantize` layers for.
        parent_modules: (:obj:`Type` | `List[Type]`): Parent module class(es) to set
            `FakeQuantize.is_fake_quant_enabled` layers in. For example, if set to
            `TransformerAttention`, only `FakeQuantize` layers in `TransformerAttention`
            modules are accessed. Defaults to `None` which does not filter.
    """
    fake_quantize_layers = get_fake_quantize_layers(
        model, parent_modules=parent_modules
    )
    previous_settings = {}
    for k, fq in fake_quantize_layers.items():
        previous_settings[k] = fq.is_fake_quant_enabled
        fq.is_fake_quant_enabled = False
    try:
        yield None
    finally:
        for k, fq in fake_quantize_layers.items():
            fq.is_fake_quant_enabled = previous_settings[k]



from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import Iterable as _Iterable
from typing import Tuple as _Tuple

import torch as _torch


def iter_named_parameters_and_buffers_with_layers(
    module: _torch.nn.Module,
) -> _Iterable[_Tuple[str, _torch.Tensor, _torch.nn.Module]]:
    """
    Generates ``(name, tensor, layer)`` tuples for all parameters and
    buffers in a :obj:`torch.nn.Module`.  Here ``tensor`` is the param or buffer,
    ``name`` is its full name in the module's state dict, and ``layer``
    is the :obj:`nn.Module` that owns ``tensor``.

    Args:
        module (:obj:`nn.Module`): The model or layer.
    """
    for layer_name, layer in module.named_modules():
        params = layer.named_parameters(
            recurse=False, remove_duplicate=False, prefix=layer_name
        )
        for name, param in list(params):
            yield name, param, layer

        buffers = layer.named_buffers(
            recurse=False, remove_duplicate=False, prefix=layer_name
        )
        for name, buffer in list(buffers):
            yield name, buffer, layer


def iter_named_parameters_and_buffers(
    module: _torch.nn.Module,
) -> _Iterable[_Tuple[str, _torch.Tensor]]:
    """
    Similar to :func:`iter_named_parameters_and_buffers_with_layers`
    but only yields names and parameters/buffers from ``module`` (so no layers).

    Args:
        module (:obj:`nn.Module`): The model or layer.
    """
    for name, tensor, _ in iter_named_parameters_and_buffers_with_layers(module):
        yield name, tensor


@_torch.no_grad()
def map_named_parameters_and_buffers(
    func: _Callable[[str, _torch.Tensor], _torch.Tensor],
    module: _torch.nn.Module,
    use_new_requires_grad_values: bool = False,
) -> _torch.nn.Module:
    """
    Overwrites all parameters and buffers in :obj:`torch.nn.Module` with the
    result of a map function, ``func``.

    Args:
        func (:obj:`callable`): A function that accepts two positional arguments,
            ``name`` and ``tensor``, and returns a new :obj:`torch.Tensor`.
            The ``tensor`` arg is a parameter or buffer from ``module``, and
            ``name`` is its full name in ``module``.  The return value is the new
            value for the parameter or buffer.
        module (:obj:`torch.nn.Module`): The model or layer with parameters/buffers
            to map.
        use_new_requires_grad_values (:obj:`bool`, optional): A flag that results
            in parameters taking the ``requires_grad`` value of the mapped tensor.
            Defaults to ``False``, in which case the ``requires_grad`` values
            do not change, regardless of the ``requires_grad`` attributes of tensors
            returned by ``func``.
    """
    cache: _Dict[int, _torch.Tensor] = {}
    for name, tensor, layer in iter_named_parameters_and_buffers_with_layers(module):
        cache_key = id(tensor)
        if cache_key in cache:
            new_tensor = cache[cache_key]  # important so that tied params stay tied
        else:
            new_tensor = func(name, tensor)
            if new_tensor is not tensor and isinstance(tensor, _torch.nn.Parameter):
                if use_new_requires_grad_values:
                    requires_grad_value = new_tensor.requires_grad
                else:
                    requires_grad_value = tensor.requires_grad
                new_tensor = _torch.nn.Parameter(
                    new_tensor, requires_grad=requires_grad_value
                )
            cache[cache_key] = new_tensor

        short_name = name.rsplit(".", maxsplit=1)[-1]
        if isinstance(new_tensor, _torch.nn.Parameter):
            delattr(layer, short_name)
            layer.register_parameter(short_name, new_tensor)
        else:
            persistent = short_name in layer.state_dict(keep_vars=True)
            delattr(layer, short_name)
            layer.register_buffer(short_name, new_tensor, persistent=persistent)
    return module



from tamm.runtime_configuration import rc as _rc

if _rc.adapters_implementation == "v1":
    from tamm._adapters_v1.utils import *  # noqa: F403
else:
    raise RuntimeError(
        f"rc.adapters_implementation '{_rc.adapters_implementation}' not recognized"
    )



import math as _math
from typing import Any as _Any
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.common import (
    MergeableLayerAdapterMixin as _MergeableLayerAdapterMixin,
)
from tamm._adapters_v1.layer_adapters.common import (
    attach_config_class as _attach_config_class,
)
from tamm._adapters_v1.layer_annotations import (
    BatchedFusedLinearProjection as _BatchedFusedLinearProjection,
)
from tamm._adapters_v1.layer_annotations import (
    BatchedLinearProjection as _BatchedLinearProjection,
)
from tamm._adapters_v1.layer_annotations import (
    FusedLinearProjection as _FusedLinearProjection,
)
from tamm._adapters_v1.layer_annotations import LinearProjection as _LinearProjection
from tamm._adapters_v1.layer_annotations import (
    get_layer_annotations as _get_layer_annotations,
)


@_attach_config_class
class LoRA(_LayerAdapter, _MergeableLayerAdapterMixin):
    _supported_layer_annotations = (_LinearProjection,)

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        rank: int,
        alpha: _Optional[float] = None,
        dropout_p: float = 0,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim

        if rank <= 0:
            raise ValueError("rank must be a positive integer.")
        self.rank = int(rank)

        self.alpha = alpha if alpha is not None else self.rank
        self.dropout_p = dropout_p

        self.a_transpose = _nn.Parameter(
            _torch.empty(input_dim, rank, device=device, dtype=dtype)
        )
        self.b_transpose = _nn.Parameter(
            _torch.empty(rank, output_dim, device=device, dtype=dtype)
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        super().reset_parameters()
        fan_in = self.a_transpose.size(0)
        _nn.init.normal_(self.a_transpose, std=_math.sqrt(1.0 / fan_in))
        _nn.init.zeros_(self.b_transpose)

    @property
    def scale(self) -> float:
        return self.alpha / self.rank

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _torch.Tensor,
    ) -> _torch.Tensor:
        if not isinstance(transformed_args, tuple):
            raise TypeError(
                f"{self.__class__.__name__} expects a tuple for transformed_args"
            )
        if len(transformed_args) != 1:
            raise ValueError(
                f"{self.__class__.__name__} expects transformed_args of length 1"
            )
        x = transformed_args[0]

        if self.training and self.dropout_p > 0:
            x = _nn.functional.dropout(x, p=self.dropout_p)
        return self._transform_outputs_impl(x, outputs)

    def _transform_outputs_impl(
        self, x: _torch.Tensor, outputs: _torch.Tensor
    ) -> _torch.Tensor:
        x = x.flatten(end_dim=-2)
        x = _torch.matmul(x, self.a_transpose)

        batch_shape = outputs.shape[:-1]
        outputs = outputs.flatten(end_dim=-2)
        outputs = _torch.addmm(outputs, x, self.b_transpose, alpha=self.scale)
        return outputs.reshape(*batch_shape, -1)

    @_torch.no_grad()
    def merge_adapter(self, wrapped_module: _nn.Module):
        wrapped_module_type = _get_layer_annotations(
            wrapped_module, filter_types=self._supported_layer_annotations
        )
        if wrapped_module_type is None:
            raise ValueError(
                f"{self.__class__.__name__} adapter's merge function only works with layers annotated "
                f"with types: {self._supported_layer_annotations}."
            )
        weight_param = wrapped_module.get_parameter(
            wrapped_module_type[0].weight_param_name
        )
        self.merge_into_weight(weight_param)

    @_torch.no_grad()
    def merge_into_weight(self, weight: _torch.Tensor) -> None:
        a_transpose = self.a_transpose.type(weight.dtype)
        b_transpose = self.b_transpose.type(weight.dtype)
        weight.addmm_(b_transpose.T, a_transpose.T, alpha=self.scale)

    def extra_repr(self) -> str:
        return (
            f"out_features={self.output_dim}, "
            f"in_features={self.input_dim}, "
            f"rank={self.rank}, "
            f"alpha={self.alpha}, "
            f"dropout_p={self.dropout_p}"
        )


@_attach_config_class
class BatchedLoRA(LoRA):
    _supported_layer_annotations = (_BatchedLinearProjection,)

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        rank: int,
        vec_dim: _List[int],
        alpha: _Optional[float] = None,
        dropout_p: float = 0,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        super().__init__(
            input_dim=input_dim,
            output_dim=output_dim,
            rank=rank,
            alpha=alpha,
            dropout_p=dropout_p,
            device=device,
            dtype=dtype,
        )
        self.vec_dim = vec_dim

        self.a_transpose = _nn.Parameter(
            _torch.empty(
                *vec_dim, self.input_dim, self.rank, device=device, dtype=dtype
            )
        )
        self.b_transpose = _nn.Parameter(
            _torch.empty(
                *vec_dim, self.rank, self.output_dim, device=device, dtype=dtype
            )
        )
        self.reset_parameters()

    def _transform_outputs_impl(
        self, x: _torch.Tensor, outputs: _torch.Tensor
    ) -> _torch.Tensor:
        return outputs + ((x @ self.a_transpose) * self.scale) @ self.b_transpose

    @_torch.no_grad()
    def merge_into_weight(self, weight: _torch.Tensor) -> None:
        weight.add_(
            (
                (self.a_transpose.to(dtype=weight.dtype) * self.scale)
                @ self.b_transpose.to(dtype=weight.dtype)
            )
        )

    def extra_repr(self) -> str:
        return (
            f"out_features={self.output_dim}, "
            f"in_features={self.input_dim}, "
            f"rank={self.rank}, "
            f"alpha={self.alpha}, "
            f"dropout_p={self.dropout_p}, "
            f"vec_dim={self.vec_dim}"
        )


class ZeroOutputModule(_nn.Module):
    def forward(self, *inputs) -> int:  # pylint: disable=unused-argument
        return 0

    def reset_parameters(self) -> None:
        pass


@_attach_config_class
class LoRAFusedMultiOutputLinear(_LayerAdapter, _MergeableLayerAdapterMixin):
    _supported_layer_annotations = (_FusedLinearProjection,)

    def __init__(
        self,
        input_dim: int,
        output_dims: _List[int],
        ranks: _List[int],
        alphas: _List[_Optional[float]] = (None, None, None),
        dropout_ps: _List[float] = (0.0, 0.0, 0.0),
        mask: _Tuple[bool, ...] = (True, False, True),
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        super().__init__()
        if not (
            len(output_dims)
            == len(ranks)
            == len(dropout_ps)
            == len(alphas)
            == len(mask)
        ):
            raise ValueError(
                f"Mismatch detected between number of output "
                f"dimensions: {len(output_dims)}, ranks: {len(ranks)}, "
                f"dropout_ps: {len(dropout_ps)}, "
                f"masks: {len(mask)} and alphas: {len(alphas)}"
            )

        self.input_dim = input_dim
        self.output_dims = output_dims
        self.mask = mask
        self.ranks = [
            rank if mask_value else None for rank, mask_value in zip(ranks, mask)
        ]
        self.alphas = [
            alpha if mask_value else None for alpha, mask_value in zip(alphas, mask)
        ]
        self.dropout_ps = [
            p if mask_value else None for p, mask_value in zip(dropout_ps, mask)
        ]
        self._create_child_lora_adapters(device=device, dtype=dtype)

    def _create_child_lora_adapters(
        self,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        for idx, mask_value in enumerate(self.mask):
            layer_name = f"lora_{idx}"
            if not mask_value:
                setattr(self, layer_name, None)
                continue

            lora = LoRA(
                input_dim=self.input_dim,
                output_dim=self.output_dims[idx],
                rank=self.ranks[idx],
                alpha=self.alphas[idx],
                dropout_p=self.dropout_ps[idx],
                device=device,
                dtype=dtype,
            )
            self.register_child_adapter(layer_name, lora)

    @property
    def num_outputs(self):
        return len(self.output_dims)

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _List[_torch.Tensor],
    ):
        if len(transformed_args) != 1:
            raise ValueError(
                f"LoRA layer for fused linear expected 1 input tensor but received "
                f"{len(transformed_args)}"
            )
        if len(outputs) != len(self.output_dims):
            raise ValueError(
                f"LoRA layer for fused linear expected f{self.num_outputs} outputs but "
                f"received {len(outputs)}"
            )
        return tuple(
            self._transform_single_output(
                lora_idx=idx,
                args=args,
                kwargs=kwargs,
                transformed_args=transformed_args,
                transformed_kwargs=transformed_kwargs,
                outputs=outputs[idx],
            )
            for idx in range(self.num_outputs)
        )

    def _transform_single_output(
        self, *, lora_idx, args, kwargs, transformed_args, transformed_kwargs, outputs
    ):
        lora = getattr(self, f"lora_{lora_idx}")
        if lora is None:
            return outputs
        return lora(
            mode="transform_outputs",
            args=args,
            kwargs=kwargs,
            transformed_args=transformed_args,
            transformed_kwargs=transformed_kwargs,
            outputs=outputs,
        )

    def _get_child_adapter_weight_params(
        self, weight_param: _torch.Tensor
    ) -> _List[_torch.Tensor]:
        return weight_param.split(self.output_dims)

    @_torch.no_grad()
    def merge_adapter(self, wrapped_module: _nn.Module):
        wrapped_module_type = _get_layer_annotations(
            wrapped_module, filter_types=self._supported_layer_annotations
        )
        if wrapped_module_type is None:
            raise ValueError(
                f"{self.__class__.__name__} adapter's merge function only works with "
                f"layers annotated with types: {self._supported_layer_annotations}."
            )
        weight_param = wrapped_module.get_parameter(
            wrapped_module_type[0].weight_param_name
        )
        weight_params = self._get_child_adapter_weight_params(weight_param)
        for idx, weight in enumerate(weight_params):
            lora = getattr(self, f"lora_{idx}")
            if lora is None:
                continue
            lora.merge_into_weight(weight)


@_attach_config_class
class BatchedLoRAFusedMultiOutputLinear(LoRAFusedMultiOutputLinear):
    _supported_layer_annotations = (_BatchedFusedLinearProjection,)

    def __init__(
        self,
        input_dim: int,
        output_dims: _List[int],
        ranks: _List[int],
        vec_dim: _List[int],
        alphas: _List[_Optional[float]] = (None, None, None),
        dropout_ps: _List[float] = (0.0, 0.0, 0.0),
        mask: _Tuple[bool, ...] = (True, False, True),
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        self.vec_dim = vec_dim
        super().__init__(
            input_dim=input_dim,
            output_dims=output_dims,
            ranks=ranks,
            alphas=alphas,
            dropout_ps=dropout_ps,
            mask=mask,
            device=device,
            dtype=dtype,
        )

    def _create_child_lora_adapters(
        self,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ):
        for idx, mask_value in enumerate(self.mask):
            layer_name = f"lora_{idx}"
            if not mask_value:
                setattr(self, layer_name, None)
                continue

            lora = BatchedLoRA(
                input_dim=self.input_dim,
                output_dim=self.output_dims[idx],
                rank=self.ranks[idx],
                vec_dim=self.vec_dim,
                alpha=self.alphas[idx],
                dropout_p=self.dropout_ps[idx],
                device=device,
                dtype=dtype,
            )
            self.register_child_adapter(layer_name, lora)

    def _get_child_adapter_weight_params(
        self, weight_param: _torch.Tensor
    ) -> _List[_torch.Tensor]:
        return weight_param.split(self.output_dims, dim=-1)



import logging as _logging
from functools import lru_cache as _lru_cache
from typing import Dict as _Dict
from typing import Optional as _Optional

from tamm import utils as _utils
from tamm.model_repo.model_repo import BaseModelRepo as _BaseModelRepo
from tamm.model_repo.publishing import PublishedModelConfig as _PublishedModelConfig

_logger = _logging.getLogger(__name__)


class RegistryModelRepo(_BaseModelRepo):
    """
    Deprecated model registry in which model configs are defined within the
    |tamm| package source.
    """

    _MODEL_CONFIG_REGISTRY = _utils.registry.Registry("Model Configs")

    def is_alive(self) -> bool:
        return True

    @_lru_cache
    def _get_configs_impl(self) -> _Dict[str, "_PublishedModelConfig"]:
        raw_configs = [
            self._MODEL_CONFIG_REGISTRY.create(config_name)
            for config_name in self._MODEL_CONFIG_REGISTRY.list()
        ]
        configs = {config.model_id: config for config in raw_configs}
        return configs

    @classmethod
    def register_model_config(
        cls,
        factory_fn=None,
        name: _Optional[str] = None,
        description: _Optional[str] = None,
    ):
        return cls._MODEL_CONFIG_REGISTRY.register(
            factory_fn, key=name, description=description
        )

    def clear_cache(self):
        ...



import functools as _functools
import importlib as _importlib
import logging as _logging
from typing import Any as _Any

from tamm import _warnings

_logger = _logging.getLogger(__name__)


class BackwardCompatibilityImporter:
    def __init__(self):
        self._registered_imports = []

    def register_backward_compatibility_import(
        self, target_module_name: str, object_name: str, source_import_path: str
    ) -> None:
        self._registered_imports.append(
            (target_module_name, object_name, source_import_path)
        )

    def execute_imports(self) -> None:
        for import_spec in self._registered_imports:
            target_module_name, object_name, source_import_path = import_spec
            try:
                obj = self._import_module_or_object(source_import_path)
            except Exception as e:
                _logger.debug(
                    "Caught exception during backward compatibility import for "
                    f"{target_module_name}.{object_name}: {e}"
                )
                continue

            if callable(obj):
                obj = _warnings.deprecate(alternative=source_import_path)(obj)

            target_module = _importlib.import_module(target_module_name)
            setattr(target_module, object_name, obj)

        del self._registered_imports[:]

    @staticmethod
    def _import_module_or_object(import_path: str) -> _Any:
        try:
            return _importlib.import_module(import_path)
        except ModuleNotFoundError:
            pass
        module_name, attribute_name = import_path.rsplit(".", 1)
        module = _importlib.import_module(module_name)
        return getattr(module, attribute_name)


@_functools.lru_cache
def get_backward_compatibility_importer() -> BackwardCompatibilityImporter:
    return BackwardCompatibilityImporter()


def register_backward_compatibility_import(
    target_module_name: str, object_name: str, source_import_path: str
) -> None:
    importer = get_backward_compatibility_importer()
    importer.register_backward_compatibility_import(
        target_module_name=target_module_name,
        object_name=object_name,
        source_import_path=source_import_path,
    )


def execute_backward_compatibility_imports() -> None:
    importer = get_backward_compatibility_importer()
    importer.execute_imports()
